#
# Copyright (c) 2020-2023 Julian Heinovski <heinovski@ccs-labs.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

import argparse
import json
import logging
import pickle
import sys
from distutils.util import strtobool
from signal import SIGINT, signal
from timeit import default_timer as timer

from plafosim import CustomFormatter, __citation__, __description__, __version__
from plafosim.algorithms import *  # noqa 401
from plafosim.simulator import DEFAULTS, Simulator
from plafosim.util import find_resource

__epilog__ = """\
Examples:
  # Configure a 100km freeway with ramps at every 10km
  plafosim --road-length 100 --ramp-interval 10

  # Configure random (normally distributed) desired driving speed of 130km/h
  plafosim --random-desired-speed true --desired-speed 36

  # Configure random trips for 500 vehicles
  plafosim --vehicles 500 --random-depart-position true --random-arrival-position true --depart-desired true

  # Pre fill the freeway with 1000 vehicles
  plafosim --vehicles 1000 --pre-fill true

  # Configure 50% of the vehicles with Advanced Cruise Control (ACC) and a headway time of 1.5s
  plafosim --penetration 0.5 --acc-headway-time 1.5

  # Enable a simple, distributed platoon formation algorithm [1] in order to form platoons every 30s
  plafosim --formation-algorithm SpeedPosition --formation-strategy distributed --execution-interval 30
"""


def format_help(parser: argparse.ArgumentParser, groups=None) -> str:
    """
    Format help message for argument groups

    Taken from https://stackoverflow.com/a/40730878.
    """

    formatter = parser._get_formatter()

    # usage
    formatter.add_usage(
        parser.usage,
        parser._actions,
        parser._mutually_exclusive_groups,
    )

    # description
    formatter.add_text(parser.description)

    if groups is None:
        groups = parser._action_groups

    # positionals, optionals and user-defined groups
    for action_group in groups:
        formatter.start_section(action_group.title)
        formatter.add_text(action_group.description)
        formatter.add_arguments(action_group._group_actions)
        formatter.end_section()

    # determine help from format above
    return formatter.format_help()


# TODO duplicated code with trace replay
def parse_args() -> (argparse.Namespace, argparse._ArgumentGroup):
    """
    Parse arguments given to this module.

    Returns
    -------
    args : argparse.Namespace
        The namespace of parsed arguments and corresponding values.
    g_gui : argparse._ArgumentGroup
        The specific argument group for the GUI properties.
    """

    # parse some parameters
    parser = argparse.ArgumentParser(
        formatter_class=CustomFormatter,
        allow_abbrev=False,
        description=__description__,
        add_help=False,
        epilog=__epilog__,
    )

    # miscellaneous
    parser.add_argument(
        '-h', '--help',
        action='store_true',
        default=argparse.SUPPRESS,
        help='show this help message and exit',
    )
    parser.add_argument(
        "-C", "--citation",
        action="version",
        help="show the citation information and exit",
        version=__citation__,
    )
    parser.add_argument(
        "-V", "--version",
        action="version",
        version=f"%(prog)s {__version__}",
    )
    parser.add_argument(
        "-d", "--default",
        action="store_true",
        help="run a simulation with the default configuration and exit",
    )
    parser.add_argument(
        "-n", "--dry-run",
        action="store_true",
        help="show the current configuration and exit",
    )
    parser.add_argument(
        "-q", "--quiet",
        action="count",
        default=0,
        help=f"The amount of verbosity levels to be removed for printing the logs to the CLI. The starting level is {logging.getLevelName(DEFAULTS['log_level'])}.",
    )
    parser.add_argument(
        "-v", "--verbosity",
        action="count",
        default=0,
        help=f"The amount of verbosity levels to be added for printing the logs to the CLI. The starting level is {logging.getLevelName(DEFAULTS['log_level'])}.",
    )
    parser.add_argument(
        "--save-snapshot",
        type=str,
        default=None,
        metavar="FILE",
        help="Save a snapshot of the scenario to FILE and exit",
    )
    parser.add_argument(
        "--load-snapshot",
        type=str,
        default=None,
        metavar="FILE",
        help="Load a snapshot of the scenario from FILE and run the simulation",
    )

    # custom help messages
    g_help = parser.add_argument_group("help messages")
    g_help.add_argument(
        "--help-all",
        action='store_true',
        default=argparse.SUPPRESS,
        help="show complete help message and exit",
    )

    # road network properties
    g_road = parser.add_argument_group("road network properties")
    g_help.add_argument(
        "--help-road",
        action='store_true',
        default=argparse.SUPPRESS,
        help="show help message for road network properties and exit",
    )
    g_road.add_argument(
        "--road-length",
        type=int,
        default=int(DEFAULTS['road_length'] / 1000),  # m -> km
        help="The length of the road in km",
    )
    g_road.add_argument(
        "--lanes",
        type=int,
        dest='number_of_lanes',
        default=DEFAULTS['lanes'],
        help="The number of lanes",
    )
    g_road.add_argument(
        "--ramp-interval",
        type=int,
        default=int(DEFAULTS['ramp_interval'] / 1000),  # m -> km
        help="The distance between any two on-/off-ramps in km",
    )
    g_road.add_argument(
        "--pre-fill",
        type=lambda x: bool(strtobool(x)),
        default=DEFAULTS['pre_fill'],
        choices=(True, False),
        help="Whether to fill the road network with vehicles using random positions and given vehicle number/density before the simulation starts",
    )

    # vehicle properties
    g_vehicles = parser.add_argument_group("vehicle properties")
    g_help.add_argument(
        "--help-vehicles",
        action='store_true',
        default=argparse.SUPPRESS,
        help="show help message for vehicle properties and exit",
    )
    g_vehicles.add_argument(
        "--vehicles",
        type=int,
        dest='number_of_vehicles',
        default=DEFAULTS['vehicles'],
        help="The (maximum) number of vehicles that are in the simulation at once. Is used for pre-fill, without a departure flow, and for some departure methods. A value of -1 disables this value.",
    )
    g_vehicles.add_argument(
        "--density",
        type=float,
        dest='vehicle_density',
        default=DEFAULTS['vehicle_density'],
        help="The (maximum) density (i.e., number of vehicles per km per lane) of vehicles that are in the simulation at once. Overrides --vehicles but behaves similarly. A value of -1 disables this value",
    )
    g_vehicles.add_argument(
        "--max-speed",
        type=float,
        default=DEFAULTS['max_speed'],
        help="The maximum possible driving speed in m/s",
    )
    g_vehicles.add_argument(
        "--acc-headway-time",
        type=float,
        default=DEFAULTS['acc_headway_time'],
        help="The headway time to be used for the ACC in s",
    )
    g_vehicles.add_argument(
        "--cacc-spacing",
        type=float,
        default=DEFAULTS['cacc_spacing'],
        help="The constant spacing to be used for the CACC in m",
    )
    g_vehicles.add_argument(
        "--penetration",
        type=float,
        dest='penetration_rate',
        default=DEFAULTS['penetration_rate'],
        help="Penetration rate of vehicles with platooning capabilities",
    )

    # trip properties
    g_trips = parser.add_argument_group("trip properties")
    g_help.add_argument(
        "--help-trips",
        action='store_true',
        default=argparse.SUPPRESS,
        help="show help message for trip properties and exit",
    )
    g_trips.add_argument(
        "--random-depart-position",
        type=lambda x: bool(strtobool(x)),
        default=DEFAULTS['random_depart_position'],
        choices=(True, False),
        help="Whether to use a random departure position for every vehicle instead of 0 m",
    )
    g_trips.add_argument(
        "--depart-all-lanes",
        type=lambda x: bool(strtobool(x)),
        default=DEFAULTS['depart_all_lanes'],
        choices=(True, False),
        help="Whether vehicles are allowed to departure on all lanes"
    )
    g_trips.add_argument(
        "--desired-speed",
        type=float,
        default=DEFAULTS['desired_speed'],
        help="The desired driving speed im m/s",
    )
    g_trips.add_argument(
        "--random-desired-speed",
        type=lambda x: bool(strtobool(x)),
        default=DEFAULTS['random_desired_speed'],
        choices=(True, False),
        help="Whether to pick a random (normally distributed) desired driving speed",
    )
    g_trips.add_argument(
        "--speed-variation",
        type=float,
        default=DEFAULTS['speed_variation'],
        help="The deviation from the desired driving speed in ratio",
    )
    g_trips.add_argument(
        "--min-desired-speed",
        type=float,
        default=DEFAULTS['min_desired_speed'],
        help="The minimum desired driving speed im m/s",
    )
    g_trips.add_argument(
        "--max-desired-speed",
        type=float,
        default=DEFAULTS['max_desired_speed'],
        help="The maximum desired driving speed im m/s",
    )
    g_trips.add_argument(
        "--random-depart-speed",
        type=lambda x: bool(strtobool(x)),
        default=DEFAULTS['random_depart_speed'],
        choices=(True, False),
        help="Whether to use a random departure speed for every vehicle instead of 0 m/s",
    )
    g_trips.add_argument(
        "--depart-desired",
        type=lambda x: bool(strtobool(x)),
        default=DEFAULTS['depart_desired'],
        choices=(True, False),
        help="Whether the vehicle should departure with its desired speed. Overrides --random-depart-speed",
    )
    g_trips.add_argument(
        "--depart-flow",
        type=lambda x: bool(strtobool(x)),
        default=DEFAULTS['depart_flow'],
        choices=(True, False),
        help="Whether to spawn vehicles in a continuous flow or as fixed number of vehicles",
    )
    g_trips.add_argument(
        "--depart-method",
        type=str,
        choices=("interval", "probability", "rate", "number"),
        default=DEFAULTS['depart_method'],
        help="The departure method of vehicles. Can be limited when depart-flow is disabled.",
    )
    g_trips.add_argument(
        "--depart-interval",
        type=float,
        default=DEFAULTS['depart_interval'],
        help="The interval between two vehicle departures in s for departure method 'interval'",
    )
    g_trips.add_argument(
        "--depart-probability",
        type=float,
        default=DEFAULTS['depart_probability'],
        help="The probability of departure per time step for departure method 'probability'",
    )
    g_trips.add_argument(
        "--depart-rate",
        type=int,
        default=DEFAULTS['depart_rate'],
        help="The rate of departure in vehicles per hour for departure method 'rate'",
    )
    g_trips.add_argument(
        "--random-arrival-position",
        type=lambda x: bool(strtobool(x)),
        default=DEFAULTS['random_arrival_position'],
        choices=(True, False),
        help="Whether to use a random arrival position for every vehicle instead of the end of the road",
    )
    g_trips.add_argument(
        "--minimum-trip-length",
        type=int,
        default=int(DEFAULTS['minimum_trip_length'] / 1000),  # m -> km
        help="The minimum trip length for a vehicle in km",
    )
    g_trips.add_argument(
        "--maximum-trip-length",
        type=int,
        default=int(DEFAULTS['maximum_trip_length'] / 1000),  # m -> km
        help="The maximum trip length for a vehicle in km",
    )

    # communication properties
    g_communication = parser.add_argument_group("communication properties")
    g_help.add_argument(
        "--help-communication",
        action='store_true',
        default=argparse.SUPPRESS,
        help="show help message for communication properties and exit",
    )
    g_communication.add_argument(
        "--communication-range",
        type=int,
        default=DEFAULTS['communication_range'],
        help="The maximum communication range between two vehicles in m. A value of -1 disables the communication range check",
    )

    # platoon properties
    g_platoon = parser.add_argument_group("platoon properties")
    g_help.add_argument(
        "--help-platoon",
        action='store_true',
        default=argparse.SUPPRESS,
        help="show help message for platoon properties and exit",
    )
    g_platoon.add_argument(
        "--start-as-platoon",
        type=lambda x: bool(strtobool(x)),
        default=DEFAULTS['start_as_platoon'],
        choices=(True, False),
        help="Whether vehicles should automatically start as one platoon",
    )
    g_platoon.add_argument(
        "--reduced-air-drag",
        type=lambda x: bool(strtobool(x)),
        default=DEFAULTS['reduced_air_drag'],
        choices=(True, False),
        help="Whether the reduced air drag due to platooning should be considered in the emissions calculation",
    )
    g_platoon.add_argument(
        "--maximum-teleport-distance",
        type=int,
        default=DEFAULTS['maximum_teleport_distance'],
        help="The maximum teleport distance in m. A value of -1 disables the check",
    )
    g_platoon.add_argument(
        "--maximum-approach-time",
        type=int,
        default=DEFAULTS['maximum_approach_time'],
        help="The maximum time for approaching a platoon during a join maneuver in s. A value of -1 disables the check",
    )
    g_platoon.add_argument(
        "--delay-teleports",
        type=lambda x: bool(strtobool(x)),
        default=DEFAULTS['delay_teleports'],
        choices=(True, False),
        help="Whether teleports (i.e., during a join maneuver) should be delayed by the time for approaching the target platoon",
    )
    g_platoon.add_argument(
        "--update-desired-speed",
        type=lambda x: bool(strtobool(x)),
        default=DEFAULTS['update_desired_speed'],
        choices=(True, False),
        help="Whether to update the platoon's desired driving speed to the average speed of all members after the formation changed",
    )

    # formation properties
    g_formation = parser.add_argument_group("formation properties")
    g_help.add_argument(
        "--help-formation",
        action='store_true',
        default=argparse.SUPPRESS,
        help="show help message for formation properties and exit",
    )
    g_formation.add_argument(
        "--formation-algorithm",
        type=str,
        default=DEFAULTS['formation_algorithm'],
        choices=globals()['algorithms'],
        help="The formation algorithm to use",
    )
    g_formation.add_argument(
        "--formation-strategy",
        type=str,
        default=DEFAULTS['formation_strategy'],
        choices=["distributed", "centralized"],
        help="The formation strategy to use",
    )
    g_formation.add_argument(
        "--execution-interval",
        type=int,
        default=DEFAULTS['execution_interval'],
        help="The interval between two iterations of a formation algorithm in s",
    )

    # infrastructure properties
    g_infrastructure = parser.add_argument_group("infrastructure properties")
    g_help.add_argument(
        "--help-infrastructure",
        action='store_true',
        default=argparse.SUPPRESS,
        help="show help message for infrastructure properties and exit",
    )
    g_infrastructure.add_argument(
        "--infrastructures",
        type=int,
        dest='number_of_infrastructures',
        default=DEFAULTS['infrastructures'],
        help="The number of infrastructures",
    )

    # simulation properties
    g_simulation = parser.add_argument_group("simulation properties")
    g_help.add_argument(
        "--help-simulation",
        action='store_true',
        default=argparse.SUPPRESS,
        help="show help message for simulation properties and exit",
    )
    g_simulation.add_argument(
        "--step-length",
        type=float,
        default=DEFAULTS['step_length'],
        help="The step length in s",
    )
    g_simulation.add_argument(
        "--time-limit",
        type=float,
        dest='max_step',
        default=float(DEFAULTS['max_step'] / 3600),  # s -> h
        help="The simulation limit in h",
    )
    g_simulation.add_argument(
        "--actions",
        type=lambda x: bool(strtobool(x)),
        default=DEFAULTS['actions'],
        choices=(True, False),
        help="Whether to enable actions",
    )
    g_simulation.add_argument(
        "--collisions",
        type=lambda x: bool(strtobool(x)),
        default=DEFAULTS['collisions'],
        choices=(True, False),
        help="Whether to enable collision checks",
    )
    g_simulation.add_argument(
        "--random-seed",
        type=int,
        default=DEFAULTS['random_seed'],
        help="The seed (>=0) for the random number generator. A value of -1 uses the current system time",
    )
    g_simulation.add_argument(
        "--progress",
        type=lambda x: bool(strtobool(x)),
        default=DEFAULTS['progress'],
        choices=(True, False),
        help="Whether to enable the (simulation) progress bar",
    )

    # GUI properties
    g_gui = parser.add_argument_group("GUI properties")
    g_help.add_argument(
        "--help-gui",
        action='store_true',
        default=argparse.SUPPRESS,
        help="show help message for GUI properties and exit",
    )
    g_gui.add_argument(
        "--gui",
        action="store_true",
        help="Enable a live SUMO GUI"
    )
    g_gui.add_argument(
        "--gui-delay",
        type=int,
        default=int(DEFAULTS['gui_delay'] * 1000),  # s -> ms
        help="The delay used in every simulation step to visualize the current network state in ms",
    )
    g_gui.add_argument(
        "--track-vehicle",
        type=int,
        dest='gui_track_vehicle',
        default=DEFAULTS['gui_track_vehicle'],
        help="The id of a vehicle to track in the GUI",
    )
    g_gui.add_argument(
        "--sumo-config",
        type=find_resource,
        default=DEFAULTS['sumo_config'],
        help="The name of the SUMO config file",
    )
    g_gui.add_argument(
        "--gui-play",
        type=lambda x: bool(strtobool(x)),
        default=DEFAULTS['gui_play'],
        choices=(True, False),
        help="Whether to start the simulation immediately",
    )
    g_gui.add_argument(
        "--gui-start",
        type=int,
        default=DEFAULTS['gui_start'],
        help="The time to connect to the GUI in s",
    )
    g_gui.add_argument(
        "--draw-ramps",
        type=lambda x: bool(strtobool(x)),
        default=DEFAULTS['draw_ramps'],
        choices=(True, False),
        help="Whether to draw on-/off-ramps",
    )
    g_gui.add_argument(
        "--draw-ramp-labels",
        type=lambda x: bool(strtobool(x)),
        default=DEFAULTS['draw_ramp_labels'],
        choices=(True, False),
        help="Whether to draw labels for on-/off-ramps",
    )
    g_gui.add_argument(
        "--draw-road-end",
        type=lambda x: bool(strtobool(x)),
        default=DEFAULTS['draw_road_end'],
        choices=(True, False),
        help="Whether to draw the end of the road",
    )
    g_gui.add_argument(
        "--draw-road-end-label",
        type=lambda x: bool(strtobool(x)),
        default=DEFAULTS['draw_road_end_label'],
        choices=(True, False),
        help="Whether to draw a label for the end of the road",
    )
    g_gui.add_argument(
        "--draw-infrastructures",
        type=lambda x: bool(strtobool(x)),
        default=DEFAULTS['draw_infrastructures'],
        choices=(True, False),
        help="Whether to draw infrastructures",
    )
    g_gui.add_argument(
        "--draw-infrastructure-labels",
        type=lambda x: bool(strtobool(x)),
        default=DEFAULTS['draw_infrastructure_labels'],
        choices=(True, False),
        help="Whether to draw labels for infrastructures",
    )
    g_gui.add_argument(
        '--screenshot-file',
        type=str,
        default=None,
        dest='screenshot_filename',
        help="The name of the screenshot file",
    )

    # result recording properties
    g_results = parser.add_argument_group("result recording properties")
    g_help.add_argument(
        "--help-results",
        action='store_true',
        default=argparse.SUPPRESS,
        help="show help message for result recording properties and exit",
    )
    g_results.add_argument(
        "--result-base-filename",
        type=str,
        default=DEFAULTS['result_base_filename'],
        help="The base filename of the result files",
    )
    g_results.add_argument(
        "--record-simulation-trace",
        type=lambda x: bool(strtobool(x)),
        default=DEFAULTS['record_simulation_trace'],
        choices=(True, False),
        help="Whether to record a continuous simulation trace",
    )
    g_results.add_argument(
        "--record-end-trace",
        type=lambda x: bool(strtobool(x)),
        default=DEFAULTS['record_end_trace'],
        choices=(True, False),
        help="Whether to record another trace item at the trip end",
    )
    g_results.add_argument(
        "--record-vehicle-trips",
        type=lambda x: bool(strtobool(x)),
        default=DEFAULTS['record_vehicle_trips'],
        choices=(True, False),
        help="Whether to record vehicle trips",
    )
    g_results.add_argument(
        "--record-vehicle-emissions",
        type=lambda x: bool(strtobool(x)),
        default=DEFAULTS['record_vehicle_emissions'],
        choices=(True, False),
        help="Whether to record vehicle emissions",
    )
    g_results.add_argument(
        "--record-vehicle-traces",
        type=lambda x: bool(strtobool(x)),
        default=DEFAULTS['record_vehicle_traces'],
        choices=(True, False),
        help="Whether to record continuous vehicles traces",
    )
    g_results.add_argument(
        "--record-vehicle-changes",
        type=lambda x: bool(strtobool(x)),
        default=DEFAULTS['record_vehicle_changes'],
        choices=(True, False),
        help="Whether to record vehicle lane changes",
    )
    g_results.add_argument(
        "--record-emission-traces",
        type=lambda x: bool(strtobool(x)),
        default=DEFAULTS['record_emission_traces'],
        choices=(True, False),
        help="Whether to record continuous emission traces",
    )
    g_results.add_argument(
        "--record-platoon-trips",
        type=lambda x: bool(strtobool(x)),
        default=DEFAULTS['record_platoon_trips'],
        choices=(True, False),
        help="Whether to record platoon trips",
    )
    g_results.add_argument(
        "--record-platoon-maneuvers",
        type=lambda x: bool(strtobool(x)),
        default=DEFAULTS['record_platoon_maneuvers'],
        choices=(True, False),
        help="Whether to record platoon maneuvers",
    )
    g_results.add_argument(
        "--record-platoon-formation",
        type=lambda x: bool(strtobool(x)),
        default=DEFAULTS['record_platoon_formation'],
        choices=(True, False),
        help="Whether to record platoon formation results",
    )
    g_results.add_argument(
        "--record-platoon-traces",
        type=lambda x: bool(strtobool(x)),
        default=DEFAULTS['record_platoon_traces'],
        choices=(True, False),
        help="Whether to record continuous platoon traces",
    )
    g_results.add_argument(
        "--record-vehicle-platoon-traces",
        type=lambda x: bool(strtobool(x)),
        default=DEFAULTS['record_vehicle_platoon_traces'],
        choices=(True, False),
        help="Whether to record continuous vehicle platoon traces",
    )
    g_results.add_argument(
        "--record-platoon-changes",
        type=lambda x: bool(strtobool(x)),
        default=DEFAULTS['record_platoon_changes'],
        choices=(True, False),
        help="Whether to record platoon lane changes",
    )
    g_results.add_argument(
        "--record-infrastructure-assignments",
        type=lambda x: bool(strtobool(x)),
        default=DEFAULTS['record_infrastructure_assignments'],
        choices=(True, False),
        help="Whether to record infrastructure assignments",
    )
    g_results.add_argument(
        "--record-vehicle-teleports",
        type=lambda x: bool(strtobool(x)),
        default=DEFAULTS['record_vehicle_teleports'],
        choices=(True, False),
        help="Whether to record vehicle teleports",
    )
    g_results.add_argument(
        "--record-prefilled",
        type=lambda x: bool(strtobool(x)),
        default=DEFAULTS['record_prefilled'],
        choices=(True, False),
        help="Whether to record results for pre-filled vehicles",
    )

    # formation algorithm specific properties
    for algorithm in globals()['algorithms']:
        globals()[f'group_{algorithm}'] = globals()[algorithm].add_parser_argument_group(parser)

        g_help.add_argument(
            f"--help-{algorithm}",
            action='store_true',
            default=argparse.SUPPRESS,
            help=f"show help message for {algorithm} formation algorithm and exit",
        )

    # print usage without any arguments
    if len(sys.argv) < 2:
        # no argument has been passed
        print(
            parser.format_usage(),
            parser.description,
            parser.epilog,
            sep='\n',
            end='',
        )
        parser.exit()

    args = parser.parse_args()

    # simple help
    misc_groups = parser._action_groups[:2]
    if 'help' in args and args.help:
        print(format_help(parser, misc_groups + [g_help]), end='')
        parser.exit()
    # complete help
    elif 'help_all' in args and args.help_all:
        print(format_help(parser), end='')
        parser.exit()
    # road network properties
    elif 'help_road' in args and args.help_road:
        print(format_help(parser, misc_groups + [g_road]), end='')
        parser.exit()
    # vehicle properties
    elif 'help_vehicles' in args and args.help_vehicles:
        print(format_help(parser, misc_groups + [g_vehicles]), end='')
        parser.exit()
    # trip properties
    elif 'help_trips' in args and args.help_trips:
        print(format_help(parser, misc_groups + [g_trips]), end='')
        parser.exit()
    # communication properties
    elif 'help_communication' in args and args.help_communication:
        print(format_help(parser, misc_groups + [g_communication]), end='')
        parser.exit()
    # platoon properties
    elif 'help_platoon' in args and args.help_platoon:
        print(format_help(parser, misc_groups + [g_platoon]), end='')
        parser.exit()
    # formation properties
    elif 'help_formation' in args and args.help_formation:
        print(format_help(parser, misc_groups + [g_formation]), end='')
        parser.exit()
    # infrastructure properties
    elif 'help_infrastructure' in args and args.help_infrastructure:
        print(format_help(parser, misc_groups + [g_infrastructure]), end='')
        parser.exit()
    # simulation properties
    elif 'help_simulation' in args and args.help_simulation:
        print(format_help(parser, misc_groups + [g_simulation]), end='')
        parser.exit()
    # GUI properties
    elif 'help_gui' in args and args.help_gui:
        print(format_help(parser, misc_groups + [g_gui]), end='')
        parser.exit()
    # result recording properties
    elif 'help_results' in args and args.help_results:
        print(format_help(parser, misc_groups + [g_results]), end='')
        parser.exit()
    # formation algorithm specific properties
    else:
        for algorithm in globals()['algorithms']:
            if f'help_{algorithm}' in args and getattr(args, f'help_{algorithm}'):
                print(format_help(parser, misc_groups + [globals()[f'group_{algorithm}']]), end='')
                parser.exit()

    # transform argument values into correct units
    args.road_length *= 1000  # km -> m
    args.ramp_interval *= 1000  # km -> m
    args.minimum_trip_length *= 1000  # km -> m
    args.maximum_trip_length *= 1000  # km -> m
    args.solver_time_limit *= 1000  # s -> ms
    args.max_step *= 3600  # h -> s
    args.gui_delay /= 1000  # ms -> s

    return args, g_gui


def load_snapshot(snapshot_filename: str) -> Simulator:
    assert snapshot_filename

    with open(snapshot_filename, "rb") as f:
        # load saved state
        simulator = pickle.load(f)
        assert isinstance(simulator, Simulator)

    return simulator


def save_snapshot(simulator: Simulator, snapshot_filename: str):
    assert isinstance(simulator, Simulator)
    assert snapshot_filename

    with open(snapshot_filename, "wb") as f:
        pickle.dump(simulator, f)


def create_simulator(**kwargs: dict) -> Simulator:
    assert kwargs

    # prepare keyword arguments for simulator
    kwargs.pop('load_snapshot')
    kwargs.pop('save_snapshot')
    kwargs['log_level'] = logging.getLevelName(max(DEFAULTS['log_level'] - ((kwargs['verbosity'] - kwargs['quiet']) * 10), 5))
    kwargs['max_step'] = int(kwargs['max_step'])

    # create new simulator
    return Simulator(**kwargs)


def main():
    """
    The main entry point of PlaFoSim.
    """

    # get argument values (and GUI argument group)
    args, g_gui = parse_args()

    if args.default:
        print("Running with default configuration...")

    if args.dry_run:
        print(f"Current configuration:\n{json.dumps(dict(sorted(vars(args).items())),indent=2)}")
        return

    simulator = None
    if args.load_snapshot:
        # load snapshot
        simulator = load_snapshot(snapshot_filename=args.load_snapshot)

        # allow to override the loaded GUI parameters
        simulator.__dict__.update(
            {
                f"_{k}": v
                for k, v in vars(args).items()
                if k in [x.dest for x in g_gui._group_actions]
            }
        )
        print(f"Loaded a snapshot of the simulation from {args.load_snapshot}. Running simulation with the loaded state...")
    else:
        # create new simulator
        simulator = create_simulator(**vars(args))

    if args.save_snapshot:
        if args.load_snapshot:
            sys.exit("ERROR: Saving a loaded snapshot does not make sense!")
        # save snapshot
        save_snapshot(simulator, snapshot_filename=args.save_snapshot)
        print(f"Saved a snapshot of the simulation to {args.save_snapshot}. Exiting...")
        return

    def handler(signal, frame):
        simulator.stop("SIGINT or CTRL-C detected! Stopping simulation...")
        exit(1)

    signal(SIGINT, handler)

    # execute simulation
    start_time = timer()

    steps = simulator.run()

    end_time = timer()
    run_time = end_time - start_time

    print(f"The simulation took {run_time} seconds ({(steps / run_time)} step/s)")


if __name__ == "__main__":
    sys.exit(main())



#
# Copyright (c) 2020-2023 Julian Heinovski <heinovski@ccs-labs.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

import sys
import time
from typing import TYPE_CHECKING

from plafosim import __version__

from .formation_algorithm import FormationAlgorithm
from .util import rgb2hex

if TYPE_CHECKING:
    from .platooning_vehicle import PlatooningVehicle  # noqa 401
    from .simulator import Simulator  # noqa 401
    from .vehicle import Vehicle  # noqa 401


def record_general_data_begin(basename: str, simulator: 'Simulator'):
    assert basename
    from .simulator import Simulator
    assert isinstance(simulator, Simulator)
    with open(f'{basename}_general.out', 'w') as f:
        f.write(f"simulation start: {time.asctime(time.localtime(time.time()))}\n")
        f.write(f"version: {__version__}\n")
        f.write(f"command:\n{' '.join(sys.argv)}\n")
        f.write(f"parameters:\n{str(simulator)}\n")


def record_general_data_end(basename: str, simulator: 'Simulator'):
    assert basename
    from .simulator import Simulator
    assert isinstance(simulator, Simulator)
    with open(f'{basename}_general.out', 'a') as f:
        f.write(f"simulation end: {time.asctime(time.localtime(time.time()))}\n")
        f.write(f"average number of vehicles: {simulator._avg_number_vehicles}\n")
        f.write(f"average number of vehicles in spawn queue: {simulator._avg_number_vehicles_queue}\n")
        f.write(f"average number of vehicles spawned: {simulator._avg_number_vehicles_spawned}\n")
        f.write(f"average number of vehicles arrived: {simulator._avg_number_vehicles_arrived}\n")
        f.write(f"average vehicle speed: {simulator._avg_vehicle_speed}\n")
        f.write(f"average number of vehicles braking rough: {simulator._avg_number_vehicles_braking_rough}\n")


def initialize_vehicle_trips(basename: str):
    assert basename
    with open(f'{basename}_vehicle_trips.csv', 'w') as f:
        f.write(
            "id,"
            "vType,"
            "eClass,"
            "vClass,"
            "depart,"
            "departLane,"
            "departPos,"
            "departSpeed,"
            "departDelay,"
            "arrival,"
            "arrivalLane,"
            "arrivalPos,"
            "arrivalSpeed,"
            "duration,"
            "routeLength,"
            "timeLoss,"
            "desiredSpeed,"
            "expectedTravelTime,"
            "travelTimeRatio,"
            "avgDrivingSpeed,"
            "avgDeviationDesiredSpeed"
            "\n"
        )


def record_vehicle_trip(
    basename: str,
    vehicle: 'Vehicle',
    depart_delay: int,
    time_loss: int,
    expected_travel_time: float,
    travel_time_ratio: float,
    average_driving_speed: float,
    average_deviation_desired_speed: float,
):
    from .vehicle import Vehicle
    assert isinstance(vehicle, Vehicle)

    with open(f'{basename}_vehicle_trips.csv', 'a') as f:
        f.write(
            f"{vehicle._vid},"
            f"{vehicle._vehicle_type.name},"
            f"{vehicle._vehicle_type.emission_class.name},"
            f"{vehicle.__class__.__name__},"
            f"{vehicle._depart_time},"
            f"{vehicle._depart_lane},"
            f"{vehicle._depart_position},"
            f"{vehicle._depart_speed},"
            f"{vehicle._depart_delay},"
            f"{vehicle._simulator.step},"
            f"{vehicle._lane},"
            f"{vehicle._position},"
            f"{vehicle._speed},"
            f"{vehicle.travel_time},"
            f"{vehicle.travel_distance},"
            f"{time_loss},"
            f"{vehicle._desired_speed},"  # use explicit individual desired speed
            f"{expected_travel_time},"
            f"{travel_time_ratio},"
            f"{average_driving_speed},"
            f"{average_deviation_desired_speed}"
            "\n"
        )


def initialize_vehicle_emissions(basename: str):
    assert basename
    with open(f'{basename}_vehicle_emissions.csv', 'w') as f:
        f.write(
            "id,"
            "CO,"
            "CO2,"
            "HC,"
            "NOx,"
            "PMx,"
            "fuel"
            "\n"
        )


def record_vehicle_emission(basename: str, vehicle: 'Vehicle'):
    assert basename
    from .vehicle import Vehicle
    assert isinstance(vehicle, Vehicle)
    with open(f'{basename}_vehicle_emissions.csv', 'a') as f:
        # TODO log estimated emissions?
        f.write(
            f"{vehicle._vid},"
            f"{vehicle._emissions['CO']},"
            f"{vehicle._emissions['CO2']},"
            f"{vehicle._emissions['HC']},"
            f"{vehicle._emissions['NOx']},"
            f"{vehicle._emissions['PMx']},"
            f"{vehicle._emissions['fuel']}"
            "\n"
        )


def initialize_platoon_trips(basename: str):
    assert basename
    with open(f'{basename}_vehicle_platoon_trips.csv', 'w') as f:
        f.write(
            "id,"
            "timeInPlatoon,"
            "distanceInPlatoon,"
            "platoonTimeRatio,"
            "platoonDistanceRatio,"
            "numberOfPlatoons,"
            "timeUntilFirstPlatoon,"
            "distanceUntilFirstPlatoon"
            "\n"
        )


def record_platoon_trip(
    basename: str,
    vehicle: 'PlatooningVehicle',
    platoon_time_ratio: float,
    platoon_distance_ratio: float,
    time_until_first_platoon: float,
    distance_until_first_platoon: float,
):
    assert basename
    from .platooning_vehicle import PlatooningVehicle
    assert isinstance(vehicle, PlatooningVehicle)

    # TODO log savings from platoon?
    with open(f'{basename}_vehicle_platoon_trips.csv', 'a') as f:
        f.write(
            f"{vehicle._vid},"
            f"{vehicle._time_in_platoon},"
            f"{vehicle._distance_in_platoon},"
            f"{platoon_time_ratio},"
            f"{platoon_distance_ratio},"
            f"{vehicle._number_platoons},"
            f"{time_until_first_platoon},"
            f"{distance_until_first_platoon}"
            "\n"
        )


def initialize_platoon_maneuvers(basename: str):
    assert basename
    with open(f'{basename}_vehicle_platoon_maneuvers.csv', 'w') as f:
        f.write(
            "id,"
            "joinsAttempted,"
            "joinsSuccessful,"
            "joinsAborted,"
            "joinsAbortedFront,"
            "joinsAbortedArbitrary,"
            "joinsAbortedRoadBegin,"
            "joinsAbortedTripBegin,"
            "joinsAbortedRoadEnd,"
            "joinsAbortedTripEnd,"
            "joinsAbortedLeaderManeuver,"
            "joinsAbortedMaxSpeed,"
            "joinsAbortedTeleportThreshold,"
            "joinsAbortedApproaching,"
            "joinsAbortedNoSpace,"
            "joinsAbortedLeaveOther,"
            "joinsFront,"
            "joinsArbitrary,"
            "joinsBack,"
            "joinsTeleportPosition,"
            "joinsTeleportLane,"
            "joinsTeleportSpeed,"
            "joinsCorrectPosition,"
            "leavesAttempted,"
            "leavesSuccessful,"
            "leavesAborted,"
            "leavesFront,"
            "leavesArbitrary,"
            "leavesBack"
            "\n"
        )


def record_vehicle_platoon_maneuvers(basename: str, vehicle: 'PlatooningVehicle'):
    assert basename
    from .platooning_vehicle import PlatooningVehicle
    assert isinstance(vehicle, PlatooningVehicle)
    with open(f'{basename}_vehicle_platoon_maneuvers.csv', 'a') as f:
        f.write(
            f"{vehicle._vid},"
            f"{vehicle._joins_attempted},"
            f"{vehicle._joins_succesful},"
            f"{vehicle._joins_aborted},"
            f"{vehicle._joins_aborted_front},"
            f"{vehicle._joins_aborted_arbitrary},"
            f"{vehicle._joins_aborted_road_begin},"
            f"{vehicle._joins_aborted_trip_begin},"
            f"{vehicle._joins_aborted_road_end},"
            f"{vehicle._joins_aborted_trip_end},"
            f"{vehicle._joins_aborted_leader_maneuver},"
            f"{vehicle._joins_aborted_max_speed},"
            f"{vehicle._joins_aborted_teleport_threshold},"
            f"{vehicle._joins_aborted_approaching},"
            f"{vehicle._joins_aborted_no_space},"
            f"{vehicle._joins_aborted_leave_other},"
            f"{vehicle._joins_front},"
            f"{vehicle._joins_arbitrary},"
            f"{vehicle._joins_back},"
            f"{vehicle._joins_teleport_position},"
            f"{vehicle._joins_teleport_lane},"
            f"{vehicle._joins_teleport_speed},"
            f"{vehicle._joins_correct_position},"
            f"{vehicle._leaves_attempted},"
            f"{vehicle._leaves_successful},"
            f"{vehicle._leaves_aborted},"
            f"{vehicle._leaves_front},"
            f"{vehicle._leaves_arbitrary},"
            f"{vehicle._leaves_back}"
            "\n"
        )


def initialize_platoon_formation(basename: str):
    assert basename
    with open(f'{basename}_vehicle_platoon_formation.csv', 'w') as f:
        f.write(
            "id,"
            "formationIterations,"
            "candidatesFound,"
            "candidatesFoundAvg,"
            "candidatesFiltered,"
            "candidatesFilteredAvg,"
            "candidatesFilteredFollower,"
            "candidatesFilteredManeuver"
            "\n"
        )


def record_platoon_formation(
    basename: str,
    vehicle: 'PlatooningVehicle',
    candidates_found_avg: float,
    candidates_filtered_avg: float
):
    assert basename
    from .platooning_vehicle import PlatooningVehicle
    assert isinstance(vehicle, PlatooningVehicle)

    with open(f'{basename}_vehicle_platoon_formation.csv', 'a') as f:
        f.write(
            f"{vehicle._vid},"
            f"{vehicle._formation_iterations},"
            f"{vehicle._candidates_found},"
            f"{candidates_found_avg},"
            f"{vehicle._candidates_filtered},"
            f"{candidates_filtered_avg},"
            f"{vehicle._candidates_filtered_follower},"
            f"{vehicle._candidates_filtered_maneuver}"
            "\n"
        )


def initialize_infrastructure_assignments(basename: str):
    assert basename
    with open(f'{basename}_infrastructure_assignments.csv', 'w') as f:
        f.write(
            "id,"
            "assignmentsSolved,"
            "assignmentsNotSolvable,"
            "assignmentsSolvedOptimal,"
            "assignmentsSolvedFeasible,"
            "assignmentsNone,"
            "assignmentsSelf,"
            "assignmentsCandidateJoinedAlready,"
            "assignmentsVehicleBecameLeader,"
            "assignmentsSuccessful"
            "\n"
        )


def record_infrastructure_assignments(basename: str, iid: int, algorithm: FormationAlgorithm):
    assert basename
    assert isinstance(algorithm, FormationAlgorithm)
    with open(f'{basename}_infrastructure_assignments.csv', 'a') as f:
        f.write(
            f"{iid},"
            f"{algorithm._assignments_solved},"
            f"{algorithm._assignments_not_solvable},"
            f"{algorithm._assignments_solved_optimal},"
            f"{algorithm._assignments_solved_feasible},"
            f"{algorithm._assignments_none},"
            f"{algorithm._assignments_self},"
            f"{algorithm._assignments_candidate_joined_already},"
            f"{algorithm._assingments_vehicle_became_leader},"
            f"{algorithm._assignments_successful}"
            "\n"
        )

# traces


def initialize_vehicle_traces(basename: str):
    assert basename
    with open(f'{basename}_vehicle_traces.csv', 'w') as f:
        f.write(
            "step,"
            "id,"
            "position,"
            "lane,"
            "speed,"
            "blocked,"
            "duration,"
            "routeLength,"
            "timeLoss,"
            "desiredSpeed,"
            "cfTargetSpeed,"
            "cfModel,"
            "color"
            "\n"
        )


def record_vehicle_trace(basename: str, step: int, vehicle: 'Vehicle'):
    assert basename
    from .vehicle import Vehicle
    assert isinstance(vehicle, Vehicle)
    with open(f'{basename}_vehicle_traces.csv', 'a') as f:
        f.write(
            f"{step},"
            f"{vehicle._vid},"
            f"{vehicle._position},"
            f"{vehicle._lane},"
            f"{vehicle._speed},"
            f"{vehicle._blocked_front},"
            f"{vehicle.travel_time},"
            f"{vehicle.travel_distance},"
            f"{vehicle._time_loss},"
            f"{vehicle.desired_speed},"  # use potential platoon desired driving speed
            f"{vehicle._cf_target_speed},"
            f"{vehicle._cf_model.name},"
            f"{rgb2hex(vehicle.color)}"  # use potential platoon color
            "\n"
        )


def initialize_vehicle_changes(basename: str):
    assert basename
    with open(f'{basename}_vehicle_changes.csv', 'w') as f:
        f.write(
            "step,"
            "id,"
            "position,"
            "from,"
            "to,"
            "speed,"
            "reason"
            "\n"
        )


def record_vehicle_change(
    basename: str,
    step: float,
    vid: int,
    position: float,
    speed: float,
    source_lane: int,
    target_lane: int,
    reason: str,
):
    with open(f'{basename}_vehicle_changes.csv', 'a') as f:
        f.write(
            f"{step},"
            f"{vid},"
            f"{position},"
            f"{source_lane},"
            f"{target_lane},"
            f"{speed},"
            f"{reason}"
            "\n"
        )


def initialize_emission_traces(basename: str):
    assert basename
    with open(f'{basename}_emission_traces.csv', 'w') as f:
        f.write(
            "step,"
            "id,"
            "CO,"
            "CO2,"
            "HC,"
            "NOx,"
            "PMx,"
            "fuel"
            "\n"
        )


def record_emission_trace_prefix(basename: str, step: float, vid: int):
    assert basename
    with open(f'{basename}_emission_traces.csv', 'a') as f:
        f.write(
            f"{step},"
            f"{vid}"
        )


def record_emission_trace_value(basename: str, value: float):
    assert basename
    with open(f'{basename}_emission_traces.csv', 'a') as f:
        f.write(
            f",{value}"
        )


def record_emission_trace_suffix(basename: str):
    assert basename
    with open(f'{basename}_emission_traces.csv', 'a') as f:
        f.write("\n")


def initialize_vehicle_platoon_traces(basename: str):
    assert basename
    with open(f'{basename}_vehicle_platoon_traces.csv', 'w') as f:
        f.write(
            "step,"
            "id,"
            "platoon,"
            "platoonRole,"
            "platoonMemberPosition,"
            "platoonDesiredSpeed,"
            "platoonSpeed"
            "\n"
        )


def record_vehicle_platoon_trace(basename: str, step: float, vehicle: 'PlatooningVehicle'):
    assert basename
    from .platooning_vehicle import PlatooningVehicle
    assert isinstance(vehicle, PlatooningVehicle)
    with open(f'{basename}_vehicle_platoon_traces.csv', 'a') as f:
        f.write(
            f"{step},"
            f"{vehicle._vid},"
            f"{vehicle._platoon.platoon_id},"
            f"{vehicle._platoon_role.name},"
            f"{vehicle._platoon.get_member_index(vehicle)},"
            f"{vehicle._platoon.desired_speed},"
            f"{vehicle._platoon.speed}"
            "\n"
        )


def initialize_platoon_traces(basename: str):
    assert basename
    with open(f'{basename}_platoon_traces.csv', 'w') as f:
        f.write(
            "step,"
            "id,"
            "leader,"
            "position,"
            "rearPosition,"
            "lane,"
            "speed,"
            "size,"
            "length,"
            "desiredSpeed"
            "\n"
        )


def record_platoon_trace(basename: str, step: float, vehicle: 'PlatooningVehicle'):
    assert basename
    from .platooning_vehicle import PlatooningVehicle
    assert isinstance(vehicle, PlatooningVehicle)
    with open(f'{basename}_platoon_traces.csv', 'a') as f:
        f.write(
            f"{step},"
            f"{vehicle._platoon.platoon_id},"
            f"{vehicle._platoon.leader.vid},"
            f"{vehicle._platoon.position},"
            f"{vehicle._platoon.rear_position},"
            f"{vehicle._platoon.lane},"
            f"{vehicle._platoon.speed},"
            f"{vehicle._platoon.size},"
            f"{vehicle._platoon.length},"
            f"{vehicle._platoon.desired_speed}"
            "\n"
        )


def initialize_platoon_changes(basename: str):
    assert basename
    with open(f'{basename}_platoon_changes.csv', 'w') as f:
        f.write(
            "step,"
            "id,"
            "position,"
            "from,"
            "to,"
            "speed,"
            "reason"
            "\n"
        )


def record_platoon_change(
    basename: str,
    step: float,
    leader: 'PlatooningVehicle',
    source_lane: int,
    target_lane: int,
    reason: str,
):
    assert basename
    from .platooning_vehicle import PlatooningVehicle
    assert isinstance(leader, PlatooningVehicle)
    from .platoon_role import PlatoonRole
    assert leader.platoon_role == PlatoonRole.LEADER and leader.platoon.leader.vid == leader.vid
    with open(f'{basename}_platoon_changes.csv', 'a') as f:
        f.write(
            f"{step},"
            f"{leader.platoon.platoon_id},"
            f"{leader.position},"
            f"{source_lane},"
            f"{target_lane},"
            f"{leader.speed},"
            f"{reason}"
            "\n"
        )


def initialize_vehicle_platoon_changes(basename: str):
    assert basename
    with open(f'{basename}_vehicle_platoon_changes.csv', 'w') as f:
        f.write(
            "step,"
            "id,"
            "platoon,"
            "position,"
            "from,"
            "to,"
            "speed,"
            "reason"
            "\n"
        )


def record_vehicle_platoon_change(
    basename: str,
    step: float,
    member: 'PlatooningVehicle',
    source_lane: int,
    target_lane: int,
    reason: str,
):
    assert basename
    from .platooning_vehicle import PlatooningVehicle
    assert isinstance(member, PlatooningVehicle)
    with open(f'{basename}_vehicle_platoon_changes.csv', 'a') as f:
        f.write(
            f"{step},"
            f"{member.vid},"
            f"{member.platoon.platoon_id},"
            f"{member.position},"
            f"{source_lane},"
            f"{target_lane},"
            f"{member.speed},"
            f"{reason}"
            "\n"
        )


def initialize_simulation_trace(basename: str):
    assert basename
    with open(f'{basename}_simulation_trace.csv', 'w') as f:
        f.write(
            "step,"
            "numberOfVehicles,"
            "vehiclesInSpawnQueue,"
            "vehiclesSpawned,"
            "vehiclesArrived,"
            "executionTime,"
            "averageVehicleSpeed,"
            "vehiclesBrakingRough"
            "\n"
        )


def record_simulation_trace(
        basename: str,
        step: float,
        vehicles_in_simulator: int,
        vehicles_in_queue: int,
        vehicles_spawned: int,
        vehicles_arrived: int,
        runtime: float,
        average_vehicle_speed: float,
        vehicles_braking_rough: int,
):
    assert basename
    with open(f'{basename}_simulation_trace.csv', 'a') as f:
        f.write(
            f"{step},"
            f"{vehicles_in_simulator},"
            f"{vehicles_in_queue},"
            f"{vehicles_spawned},"
            f"{vehicles_arrived},"
            f"{runtime},"
            f"{average_vehicle_speed},"
            f"{vehicles_braking_rough}"
            "\n"
        )


def initialize_vehicle_teleports(basename: str):
    assert basename
    with open(f'{basename}_vehicle_teleports.csv', 'w') as f:
        f.write(
            "step,"
            "id,"
            "oldPosition,"
            "oldLane,"
            "oldSpeed,"
            "newPosition,"
            "newLane,"
            "newSpeed"
            "\n"
        )


def record_vehicle_teleport(
    basename: str,
    step: float,
    vid: int,
    old_position: float,
    old_lane: int,
    old_speed: float,
    new_position: float,
    new_lane: int,
    new_speed: float,
):
    assert basename
    with open(f'{basename}_vehicle_teleports.csv', 'a') as f:
        f.write(
            f"{step},"
            f"{vid},"
            f"{old_position},"
            f"{old_lane},"
            f"{old_speed},"
            f"{new_position},"
            f"{new_lane},"
            f"{new_speed}"
            "\n"
        )



#
# Copyright (c) 2020-2023 Julian Heinovski <heinovski@ccs-labs.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

import random


def get_desired_speed(
    desired_speed: float,
    rng: random.Random,
    speed_variation: float,
    min_desired_speed: float,
    max_desired_speed: float,
    random_desired_speed: bool = False,
) -> float:
    """
    Return a (random) departure speed.
    """

    if random_desired_speed:
        # normal distribution
        speed = desired_speed * rng.normalvariate(1.0, speed_variation)
        # TODO new dice roll instead of cutting?
        speed = max(speed, min_desired_speed)
        speed = min(speed, max_desired_speed)
    else:
        speed = desired_speed

    return speed


def get_depart_speed(
    desired_speed: float,
    rng: random.Random,
    depart_desired: bool = False,
    random_depart_speed: bool = False,
) -> float:
    """
    Return a (random) departure speed.

    Parameters
    ----------
    desired_speed : float
        The desired speed to consider
    """

    if random_depart_speed:
        # make sure to also include the desired speed itself
        depart_speed = rng.randrange(0, desired_speed + 1, 1)
    else:
        depart_speed = 0

    if depart_desired:
        depart_speed = desired_speed

    return depart_speed


def get_arrival_position(
    depart_position: int,
    road_length: int,
    ramp_interval: int,
    min_trip_length: int,
    max_trip_length: int,
    rng: random.Random,
    random_arrival_position: bool = False,
    pre_fill: bool = False,
) -> int:
    """
    Return a (random) arrival position for a given departure position.

    This considers the ramp interval, road length, and minimum trip length.

    Parameters
    ----------
    depart_position : int
        The departure position to consider
    road_length : int
        The length of the entire road
    ramp_interval : int
        The distance between two on-/off-ramps
    min_trip_length : int
        The minimum trip length
    max_trip_length : int
        The maximum trip length
    rng : random.Random
        The RNG to be used
    random_arrival_position : bool
        Whether to use random arrival positions
    prefill : bool, optional
        Whether the trip is for a pre-filled vehicle
    """

    # set minimum theoretical arrival position
    if pre_fill:
        # We cannot use the minimum trip time here,
        # since the pre-generation is supposed to produce a snapshot of a realistic simulation.
        # But we can assume that a vehicle has to drive at least 1m
        assert depart_position <= (road_length - 1)
        min_arrival = depart_position + 1
        max_arrival = min(depart_position + max_trip_length - ramp_interval, road_length)
    else:
        # make sure that the vehicles drive at least for the minimum length of a trip
        # and at least for one ramp
        min_arrival = min(depart_position + max(min_trip_length, 1), road_length)
        max_arrival = min(depart_position + max_trip_length, road_length)
    min_arrival_ramp = min_arrival + (ramp_interval - min_arrival) % ramp_interval
    max_arrival_ramp = max_arrival + (ramp_interval - max_arrival) % ramp_interval
    assert min_arrival_ramp >= 0
    assert min_arrival_ramp <= road_length
    assert min_arrival_ramp <= max_arrival_ramp
    assert max_arrival_ramp <= road_length
    if min_arrival % ramp_interval == 0:
        assert min_arrival == min_arrival_ramp

    if random_arrival_position:
        # make sure to also include the end of the road itself
        arrival_position = rng.randrange(min_arrival_ramp, max_arrival_ramp + 1, ramp_interval)
        assert min_arrival_ramp <= arrival_position <= max_arrival_ramp
        assert arrival_position <= road_length
    else:
        # simply drive until the end of the trip or the road
        arrival_position = min(max_arrival_ramp, road_length)

    assert arrival_position <= road_length
    assert arrival_position > depart_position
    assert arrival_position - depart_position <= max_trip_length

    return arrival_position



#
# Copyright (c) 2020-2023 Julian Heinovski <heinovski@ccs-labs.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

import os
from importlib import import_module
from inspect import isclass
from pkgutil import iter_modules

from ..formation_algorithm import FormationAlgorithm

# iterate through the modules in the current package
globals()['algorithms'] = list()
# based on https://julienharbulot.com/python-dynamical-import.html
package_dir = os.path.dirname(__file__)
for (_, module_name, _) in iter_modules([package_dir]):
    # import the module and iterate through its attributes
    module = import_module(f"{__name__}.{module_name}")
    for attribute_name in dir(module):
        attribute = getattr(module, attribute_name)

        if isclass(attribute) and issubclass(attribute, FormationAlgorithm) and attribute != FormationAlgorithm:
            # Add the class to this package's variables
            globals()[attribute_name] = attribute
            globals()['algorithms'].append(attribute_name)



#
# Copyright (c) 2020-2023 Julian Heinovski <heinovski@ccs-labs.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

import logging
import math
import sys
from typing import TYPE_CHECKING

from .algorithms import *  # noqa 401
from .gui import change_gui_vehicle_color
from .mobility import CF_Model, is_gap_safe
from .platoon import Platoon
from .platoon_role import PlatoonRole
from .statistics import (
    record_platoon_formation,
    record_platoon_trace,
    record_platoon_trip,
    record_vehicle_change,
    record_vehicle_platoon_maneuvers,
    record_vehicle_platoon_trace,
    record_vehicle_teleport,
)
from .util import round_to_next_base
from .vehicle import Vehicle
from .vehicle_type import VehicleType

if TYPE_CHECKING:
    from .simulator import Simulator  # noqa 401

LOG = logging.getLogger(__name__)


class PlatooningVehicle(Vehicle):
    """
    A vehicle that has platooning functionality.
    """

    def __init__(
            self,
            simulator: 'Simulator',
            vid: int,
            vehicle_type: VehicleType,
            depart_position: int,
            arrival_position: int,
            desired_speed: float,
            depart_lane: int,
            depart_speed: float,
            depart_time: float,
            depart_delay: float = 0,
            communication_range: int = 1000,  # TODO use defaults
            acc_headway_time: float = 1.0,  # TODO use defaults
            cacc_spacing: float = 5.0,  # TODO use defaults
            formation_algorithm: str = None,
            execution_interval: int = 1,  # TODO use defaults
            pre_filled: bool = False,
            **kw_args,
    ):
        # TODO use global default values
        """
        Initialize a platooning vehicle instance.

        Parameters
        ----------
        simulator : Simulator
            The global simulator object
        vid : int
            The id of the vehicle
        vehicle_type : VehicleType
            The vehicle type of the vehicle
        depart_position : int
            The departure position of the vehicle
        arrival_position : int
            The arrival position of the vehicle
        desired_speed : float
            The desired driving speed of the vehicle
        depart_lane : int
            The departure lane of the vehicle
        depart_speed : float
            The departure speed of the vehicle
        depart_time : float
            The actual departure time of the vehicle
        depart_delay : float
            The time the vehicle had to wait before starting its trip
        communication_range : int
            The maximum communication range of the vehicle
        acc_headway_time: float
            The headway time for the ACC
        cacc_spacing: float
            The constant spacing for the CACC
        formation_algorithm: str
            The platoon formation algorithm to use
        execution_interval: int
            The interval for executing the formation algorithm
        pre_filled : bool
            Whether this vehicle was pre-filled
        """

        super().__init__(
            simulator,
            vid,
            vehicle_type,
            depart_position,
            arrival_position,
            desired_speed,
            depart_lane,
            depart_speed,
            depart_time,
            depart_delay,
            communication_range,
            pre_filled,
        )

        self._cf_model = CF_Model.ACC
        self._acc_headway_time = acc_headway_time
        self._acc_lambda = 0.1  # see Eq. 6.18 of R. Rajamani, Vehicle Dynamics and Control, 2nd. Springer, 2012.
        self._cacc_spacing = cacc_spacing
        self._platoon_role = PlatoonRole.NONE  # the current platoon role
        self._platoon = Platoon(vid, [self], desired_speed)
        self._in_maneuver = False
        # for a JOINER
        self._join_approach_step = None
        self._join_data_leader = None
        self._join_data_last = None
        self._join_data_new_position = None
        # for a LEADER
        self._joiner = None

        if formation_algorithm:
            # initialize formation algorithm
            try:
                self._formation_algorithm = globals()[formation_algorithm](self, **kw_args)
            except KeyError:
                sys.exit(f"ERROR: Unknown formation algorithm {formation_algorithm}! Did you import it?")
            self._execution_interval = execution_interval

            # initialize timers
            if pre_filled:
                self._last_formation_step = simulator._rng.randint(0, self._execution_interval - 1)
            else:
                self._last_formation_step = self._depart_time  # initialize with vehicle start
            self._last_advertisement_step = None

        else:
            self._formation_algorithm = None

        ## statistics

        # platoon statistics
        self._first_platoon_join_time = -1
        self._last_platoon_join_time = -1
        self._time_in_platoon = 0
        self._first_platoon_join_position = -1
        self._last_platoon_join_position = -1
        self._distance_in_platoon = 0
        self._number_platoons = 0

        # maneuver statistics
        self._joins_attempted = 0
        self._joins_succesful = 0
        self._joins_aborted = 0
        self._joins_aborted_front = 0
        self._joins_aborted_arbitrary = 0
        self._joins_aborted_road_begin = 0
        self._joins_aborted_trip_begin = 0
        self._joins_aborted_road_end = 0
        self._joins_aborted_trip_end = 0
        self._joins_aborted_leader_maneuver = 0
        self._joins_aborted_max_speed = 0
        self._joins_aborted_teleport_threshold = 0
        self._joins_aborted_approaching = 0
        self._joins_aborted_no_space = 0
        self._joins_aborted_leave_other = 0

        self._joins_front = 0
        self._joins_arbitrary = 0
        self._joins_back = 0
        self._joins_teleport_position = 0
        self._joins_teleport_lane = 0
        self._joins_teleport_speed = 0
        self._joins_correct_position = 0
        self._leaves_attempted = 0
        self._leaves_successful = 0
        self._leaves_aborted = 0
        self._leaves_front = 0
        self._leaves_arbitrary = 0
        self._leaves_back = 0

        # formation statistics
        self._formation_iterations = 0
        self._candidates_found = 0
        self._candidates_filtered = 0
        self._candidates_filtered_follower = 0
        self._candidates_filtered_maneuver = 0

    @property
    def acc_headway_time(self) -> float:
        """
        Return the ACC headway time of the vehicle.
        """

        return self._acc_headway_time

    @property
    def desired_headway_time(self) -> float:
        """
        Return the desired headway time of the vehicle.

        This is based on the currently active car following model.
        """

        if self._cf_model == CF_Model.ACC:
            return self._acc_headway_time
        elif self._cf_model == CF_Model.CACC:
            return 0
        return super().desired_headway_time

    @property
    def desired_speed(self) -> float:
        """
        Return the desired driving speed of the vehicle.

        If the vehicle is in a platoon, it returns the desired driving speed of the entire platoon.
        """

        return self._platoon.desired_speed if self.is_in_platoon() else self._desired_speed

    @property
    def desired_gap(self) -> float:
        """
        Return the desired gap of the vehicle.

        This is based on the currently active car following model.
        """
        if self._cf_model == CF_Model.CACC:
            return self._cacc_spacing
        return super().desired_gap

    @property
    def platoon_role(self) -> PlatoonRole:
        """
        Return the current platoon role of the vehicle.
        """

        return self._platoon_role

    @property
    def platoon(self) -> Platoon:
        """
        Return the platoon of the vehicle.
        """

        return self._platoon

    def is_in_platoon(self) -> bool:
        """
        Return whether the vehicle currently is in a platoon.

        This is based on the current PlatoonRole.
        A joining or leaving vehicle is either not yet or still part of a platoon, thus the returned value should be true.
        """

        return self._platoon_role is not PlatoonRole.NONE

    def get_front_gap(self) -> float:
        """
        Return the gap to the vehicle in the front.

        This imitates a measurement of the front radar sensor.
        """

        return self._simulator._get_predecessor_rear_position(self) - self._position

    def get_front_speed(self) -> float:
        """
        Return the speed to the vehicle in the front.

        This imitates a measurement of the front radar sensor.
        """

        return self._simulator._get_predecessor_speed(self)

    @property
    def in_maneuver(self) -> bool:
        """
        Return whether the vehicle is currently in a maneuver.
        """

        return self._in_maneuver

    @in_maneuver.setter
    def in_maneuver(self, var: bool):
        """
        Set the maneuver status of the vehicle.

        Parameters
        ----------
        var : bool
            The maneuver status
        """

        assert var != self._in_maneuver, f"Maneuver request {var} does not match maneuver status {self._in_maneuver} of {self._vid}!"
        self._in_maneuver = var

    @property
    def time_in_platoon(self) -> int:
        """
        Return the travelled time within platoons.
        """

        return self._time_in_platoon

    @property
    def distance_in_platoon(self) -> float:
        """
        Return the travelled distance within platoons.
        """

        return self._distance_in_platoon

    @property
    def color(self) -> tuple:
        """Return the current color of the vehicle."""

        return self._platoon.leader._color if self.is_in_platoon() else self._color

    def _calculate_emission(self, a: float, v: float, f: list, scale: float) -> float:
        """
        Calculate the emitted pollutant amount using the given speed and acceleration.

        Parameters
        ----------
        a : float
            The current acceleration
        v : float
            The current speed
        f : list
            The emission factors to use for current emission variable to be calculated
        scale : float
            The scale to normalize the calculated value
        """

        # calculate impact of platooning on air drag based on
        # Charles-Henri Bruneau, Khodor Khadra and Iraj Mortazavi,
        # "Flow analysis of square-back simplified vehicles in platoon,"
        # International Journal of Heat and Fluid Flow, vol. 66, pp. 43–59,
        # August 2017. Table 5: d = L, vehicle length = 5m, distance = 5m
        air_drag_change = 0.0
        if self._platoon_role == PlatoonRole.LEADER:
            # savings by followers
            assert self._platoon.size > 1
            air_drag_change = 0.12
        elif self._platoon_role == PlatoonRole.FOLLOWER:
            # savings by leader/front vehicles
            if self is self._platoon.last:
                # last vehicle
                air_drag_change = 0.23
            else:
                # in between
                air_drag_change = 0.27

        # calculate impact of air drag on emissions based on
        # Gino Sovran, "Tractive-Energy-Based Formulae for the Impact of
        # Aerodynamics on Fuel Economy Over the EPA Driving Schedules,"
        # SAE International, Technical Paper, 830304, February 1983.
        if self._simulator._reduced_air_drag:
            emission_change = air_drag_change * 0.46
        else:
            emission_change = 0.0

        return super()._calculate_emission(a, v, f, scale) * (1.0 - emission_change)

    def finish(self):
        """
        Clean up the instance of the PlatooningVehicle.

        This includes leaving the platoon and mostly statistics recording.
        """

        if (self._position < self._arrival_position):
            LOG.warning(f"{self._vid}'s finish method was called even though vehicle did not arrive yet!")
            return

        super().finish()

        # clean up platoon
        if self.is_in_platoon():
            self._leave()

        assert self.travel_time > 0
        assert self._time_in_platoon >= 0
        platoon_time_ratio = self._time_in_platoon / self.travel_time
        assert self.travel_distance > 0
        assert self._distance_in_platoon >= 0
        platoon_distance_ratio = self._distance_in_platoon / self.travel_distance

        LOG.info(f"{self._vid} drove {self._time_in_platoon}s ({self._distance_in_platoon}m) in a platoon, {platoon_time_ratio * 100}% ({platoon_distance_ratio * 100}%) of the trip")

        # statistic recording

        # TODO use pre_filled flag
        if not self._simulator._record_prefilled and self._depart_time == -1:
            # we do not record statistics for pre-filled vehicles
            LOG.debug(f"Not recording statistics for pre-filled vehicle {self._vid}")
            return

        candidates_found_avg = (
            self._candidates_found / self._formation_iterations
            if self._formation_iterations > 0
            else 0
        )
        candidates_filtered_avg = (
            self._candidates_filtered / self._formation_iterations
            if self._formation_iterations > 0
            else 0
        )

        assert platoon_time_ratio >= 0
        assert platoon_distance_ratio >= 0

        if self._first_platoon_join_time == -1:
            # was not set yet (no platoon)
            assert self._first_platoon_join_position == -1
            time_until_first_platoon = None
            distance_until_first_platoon = None
        else:
            # NOTE: this produces wrong values when pre-filled
            time_until_first_platoon = self._first_platoon_join_time - self._depart_time
            assert time_until_first_platoon >= 0
            assert self._first_platoon_join_position != -1
            # NOTE: this produces wrong values when pre-filled
            distance_until_first_platoon = self._first_platoon_join_position - self._depart_position
            assert distance_until_first_platoon >= 0

        if self._simulator._record_platoon_trips:
            record_platoon_trip(
                basename=self._simulator._result_base_filename,
                vehicle=self,
                platoon_time_ratio=platoon_time_ratio,
                platoon_distance_ratio=platoon_distance_ratio,
                time_until_first_platoon=time_until_first_platoon,
                distance_until_first_platoon=distance_until_first_platoon,
            )

        if self._simulator._record_platoon_maneuvers:
            record_vehicle_platoon_maneuvers(basename=self._simulator._result_base_filename, vehicle=self)

        if self._simulator._record_platoon_formation:
            record_platoon_formation(
                basename=self._simulator._result_base_filename,
                vehicle=self,
                candidates_found_avg=candidates_found_avg,
                candidates_filtered_avg=candidates_filtered_avg,
            )

    def _action(self, step: float):
        """
        Triggers specific actions of a PlatooningVehicle.

        Parameters
        ----------
        step : float
            The current simulation step
        """

        super()._action(step)

        # continue jojn maneuver (approaching the target platoon)
        if self._join_approach_step:
            assert self._in_maneuver
            if step >= self._join_approach_step:
                self._join_teleport(self._join_data_leader, self._join_data_last, self._join_data_new_position)
                # cleaning up
                self._join_approach_step = None
                self._join_data_leader._joiner = None
                self._join_data_leader = None
                self._join_data_last = None
                self._join_data_new_position = None

        # do platoon formation
        if self._formation_algorithm:
            # execute formation algorithm at every execution interval
            if step >= self._last_formation_step + self._execution_interval:
                # search for a platoon (depending on the algorithm)
                self._formation_algorithm.do_formation()
                self._last_formation_step = step

    def info(self) -> str:
        """
        Return information about the PlatooningVehicle.
        """

        return f"{super().info()}, platoon {self._platoon}"

    def __str__(self) -> str:
        """
        Return the str representation of the platooning vehicle.
        """

        self_dict = self.__dict__.copy()
        self_dict.update({'_vehicle_type': str(self._vehicle_type)})  # use str representation of vehicle type
        self_dict.update({'_platoon': str(self._platoon)})  # use str representation of platoon
        return str(self_dict)

    def _statistics(self):
        """
        Write continuous statistics.
        """

        super()._statistics()

        if not self._simulator._record_prefilled and self._depart_time == -1:
            # we do not record statistics for pre-filled vehicles
            return

        if self._simulator._record_platoon_traces:
            # write statistics about the current platoon
            if self._platoon.leader is self:
                # we do not want followers to record that
                assert self._platoon_role != PlatoonRole.FOLLOWER
                record_platoon_trace(
                    basename=self._simulator._result_base_filename,
                    step=self._simulator.step,
                    vehicle=self,
                )

        if self._simulator._record_vehicle_platoon_traces:
            # write statistics about this vehicle within the current platoon
            record_vehicle_platoon_trace(
                basename=self._simulator._result_base_filename,
                step=self._simulator.step,
                vehicle=self,
            )

    # TODO rework to only include "neighbors" and move platoon extraction to formation algorithm
    def _get_available_platoons(self):
        """
        Return the available platoon candidates of the vehicle.

        This imitates neighbor maintenance by using a neighbor table.
        """

        # HACK FOR AVOIDING MAINTAINING A NEIGHBOR TABLE (for now)
        platoons = []
        for vehicle in self._simulator._vehicles.values():

            # filter own vehicle
            if vehicle is self:
                continue

            # filter non-available vehicles which are technically not able to do platooning
            if not isinstance(vehicle, PlatooningVehicle):
                continue

            # filter non-platoons
            if vehicle.platoon_role != PlatoonRole.LEADER and vehicle.platoon_role != PlatoonRole.NONE:
                continue

            # filter non-available vehicles based on communication range
            if abs(vehicle.position - self._position) > self._communication_range:
                LOG.trace(f"{self._vid}'s neighbor {vehicle.vid} is out of communication range ({self._communication_range}m)")
                continue

            platoons.append(vehicle.platoon)

        return platoons

    def _join(self, platoon_id: int, leader_id: int):
        """
        Lets a vehicle join a platoon.

        Communication and fine-grained maneuver control is out-of-scope and thus omitted.

        Parameters
        ----------
        platoon_id : int
            The id of the target platoon
        leader_id : int
            The id of the leader of the target platoon
        """

        assert not self.is_in_platoon()

        LOG.trace(f"{self._vid} is trying to join platoon {platoon_id} (leader {leader_id})")
        self._joins_attempted += 1

        assert not self.in_maneuver
        self.in_maneuver = True
        self._platoon_role = PlatoonRole.JOINER

        leader = self._simulator._vehicles[leader_id]
        assert isinstance(leader, PlatooningVehicle)

        # correct platoon
        assert leader.platoon.platoon_id == platoon_id
        # correct leader of that platoon
        assert leader.vid == leader.platoon.leader.vid

        if leader.in_maneuver:
            LOG.warning(f"{self._vid}'s new leader {leader_id} was already in a maneuver! Aborting the join maneuver!")
            self.in_maneuver = False
            self._platoon_role = PlatoonRole.NONE

            self._joins_aborted += 1
            self._joins_aborted_leader_maneuver += 1
            return

        # HACK for determining the join position
        if self._position > leader.platoon.position:
            # TODO join at front
            LOG.debug(f"{self._vid} is in front of the target platoon {platoon_id} ({leader_id})")
            LOG.warning("Join at the front of a platoon is not yet implemented! Aborting the join maneuver!")
            self.in_maneuver = False
            self._platoon_role = PlatoonRole.NONE

            self._joins_front += 1
            self._joins_aborted += 1
            self._joins_aborted_front += 1
            return
        elif self._position > leader.platoon.last.position:
            # TODO join at (arbitrary position) in the middle
            LOG.debug(f"{self._vid} is in front of (at least) the last vehicle {leader.platoon.last.vid} of the target platoon {platoon_id} ({leader_id})")
            LOG.warning("Join at arbitrary positions of a platoon is not yet implemented! Aborting the join maneuver!")
            self.in_maneuver = False
            self._platoon_role = PlatoonRole.NONE

            self._joins_arbitrary += 1
            self._joins_aborted += 1
            self._joins_aborted_arbitrary += 1
            return

        # join at back
        self._joins_back += 1

        last = leader.platoon.last
        new_position = last.rear_position - self._cacc_spacing
        LOG.trace(f"{self._vid}'s new position is ({new_position},{leader.platoon.lane}) (current {self._position},{self._lane})")

        if new_position < self.length:
            # we cannot join since we would be outside of the road
            LOG.warning(f"{self._vid}'s new position would be too close to the beginning of the road! Aborting the join maneuver!")
            self.in_maneuver = False
            self._platoon_role = PlatoonRole.NONE

            self._joins_aborted += 1
            self._joins_aborted_road_begin += 1
            return

        if new_position < self._depart_position:
            LOG.warning(f"{self._vid}'s new position would be before its departure position! Aborting the join maneuver!")
            self.in_maneuver = False
            self._platoon_role = PlatoonRole.NONE

            self._joins_aborted += 1
            self._joins_aborted_trip_begin += 1
            return
        # TODO should we also avoid teleporting backwards at all?
        # for now this is allowed to simulate decelerating and waiting for the platoon for passing further

        if new_position >= self._simulator.road_length:
            LOG.warning(f"{self._vid}'s new position would be outside of the road! Aborting the join maneuver")
            self.in_maneuver = False
            self._platoon_role = PlatoonRole.NONE

            self._joins_aborted += 1
            self._joins_aborted_road_end += 1
            return

        if new_position >= self._arrival_position:
            LOG.warning(f"{self._vid}'s new position would be outside of its trip! Aborting the join maneuver")
            self.in_maneuver = False
            self._platoon_role = PlatoonRole.NONE

            self._joins_aborted += 1
            self._joins_aborted_trip_end += 1
            return

        if self.max_speed <= leader.platoon.speed:
            # we will never be able to approach the platoon or maintain the platoon speed
            LOG.warning(f"{self._vid}'s maximum speed is too low such that it can never reach the platoon! Aborting the join maneuver!")
            self.in_maneuver = False
            self._platoon_role = PlatoonRole.NONE

            self._joins_aborted += 1
            self._joins_aborted_max_speed += 1
            return

        if abs(new_position - self._position) > self._simulator._maximum_teleport_distance:
            # the vehicle is too far away to be teleported
            LOG.warning(f"{self._vid} is too far away from the target platoon to realistically do a teleport! Aborting the join maneuver!")
            self.in_maneuver = False
            self._platoon_role = PlatoonRole.NONE

            self._joins_aborted += 1
            self._joins_aborted_teleport_threshold += 1
            return

        assert new_position >= self.length
        assert new_position <= self._simulator.road_length

        # consider the actual approaching duration
        total_approach_time = self.calculate_approaching_time(new_position, leader.platoon.speed)

        assert total_approach_time >= 0
        if total_approach_time > self._simulator._maximum_appraoch_time:
            # approaching the platoon would take too long
            LOG.warning(f"It would take too long ({total_approach_time}s) for {self._vid} to approach the platoon {leader.platoon.platoon_id} ({leader.vid})! Aborting the join maneuver!")
            self.in_maneuver = False
            self._platoon_role = PlatoonRole.NONE

            self._joins_aborted += 1
            self._joins_aborted_approaching += 1
            return

        # the join has been "allowed" by the leader
        # the actual join procedure starts here
        assert not leader.in_maneuver
        leader.in_maneuver = True
        assert leader.cf_model == CF_Model.ACC

        # delay teleport by approach duration
        if self._simulator._delay_teleports and total_approach_time > 0:
            # round approach time to next simulation step
            total_approach_time = round_to_next_base(total_approach_time, self._simulator.step_length)
            # the platoon will be driving while we are waiting
            new_position += total_approach_time * leader.platoon.speed
            # FIXME the extra distance from above increases the distance to approach, which should also increase the total approaching time (by time to drive this extra distance with maximum speed)
            # TODO use classical formulas from school physics: when do 2 cars meet and where?
            if new_position >= self._simulator.road_length:
                LOG.warning(f"{self._vid}'s new position would be outside of the road! Aborting the join maneuver")
                self.in_maneuver = False
                self._platoon_role = PlatoonRole.NONE
                leader.in_maneuver = False

                self._joins_aborted += 1
                self._joins_aborted_road_end += 1
                return

            # check if the new position would be outside of the trip
            # NOTE: we need to add the distance the vehicle drives within one step as well in order to avoid that the vehicles finishes within the next step after the maneuver
            if new_position + self._simulator.step_length * self.speed >= self._arrival_position:
                LOG.warning(f"{self._vid}'s new position would be outside of its trip! Aborting the join maneuver")
                self.in_maneuver = False
                self._platoon_role = PlatoonRole.NONE
                leader.in_maneuver = False

                self._joins_aborted += 1
                self._joins_aborted_trip_end += 1
                return

            self._join_approach_step = self._simulator.step + total_approach_time
            self._join_data_leader = leader
            self._join_data_last = last
            self._join_data_new_position = new_position
            leader._joiner = self
            LOG.trace(f"Scheduled the teleport of vehicle {self._vid} to step {self._join_approach_step} (new position: {new_position})")
        else:
            # perform the teleport now
            LOG.trace(f"The teleport for vehicle {self._vid} will be performed instantaneous.")
            self._join_teleport(leader, last, new_position)

    def _join_teleport(self, leader: 'PlatooningVehicle', last: 'PlatooningVehicle', new_position: float):
        """
        Parameters
        ----------
        leader : PlatooningVehicle
            The leader of the target platoon
        last : PlatooningVehicle
            The last vehicle of the target platoon
        new_position : float
            The new position for joining the target platoon
        """

        LOG.trace(f"Continuing the join maneuver for vehicle {self._vid} -> {leader.platoon.platoon_id} ({leader.vid})")

        if leader.position >= leader.arrival_position:
            # the leader is gone (or will be gone soon)
            # TODO we might still want to teleport the vehicle for realism
            LOG.warning(f"{self._vid}'s platoon leader meanwhile reached the end of its trip! Aborting the join maneuver")
            self.in_maneuver = False
            self._platoon_role = PlatoonRole.NONE
            leader.in_maneuver = False

            self._joins_aborted += 1
            self._joins_aborted_trip_end += 1
            return

        assert self._in_maneuver
        assert self._platoon_role is PlatoonRole.JOINER
        assert leader.in_maneuver
        assert (
            leader.platoon_role is PlatoonRole.LEADER
            or leader.platoon_role is PlatoonRole.NONE
        )

        # re-calculate new position
        new_position = last.rear_position - self._cacc_spacing
        LOG.trace(f"{self._vid}'s new position is ({new_position},{leader.platoon.lane}) (current {self._position},{self._lane})")

        assert new_position >= self.length
        assert new_position >= self._depart_position
        assert new_position <= self._simulator.road_length
        assert new_position <= self._arrival_position

        platoon_successor = self._simulator._get_successor(last)
        if not platoon_successor or platoon_successor is self:
            LOG.trace(f"{self._vid}'s new position is {new_position} (behind {last.vid})")
        else:
            LOG.trace(f"{self._vid}'s new position is {new_position} (between {last.vid} and {platoon_successor.vid})")

            # MAKE SPACE FOR THE JOINER
            # the idea is to move the vehicle(s) behind the platoon to make room for the joiner
            # we start with the first one (direct follower) and proceed with the next vehicle(s) as long as we need more space
            # thus, we move only as little vehicles and as little as necessary (until they reach the minimum gap to the front vehicle)
            # this way, it is more realistic and keeps the correct order of vehicles without producing an incorrect state during the process
            # we might also move the method to the vehicle class or make it even part of the simulator

            # how big needs the gap behind the platoon in front of the platoon successor to be
            # CACC spacing + joiner + successor's min gap
            required_gap = self._cacc_spacing + self.length + platoon_successor.min_gap
            LOG.trace(f"We need a gap of {required_gap}m behind the platoon (vehicle {last.vid}) to teleport vehicle {self._vid}")
            current_gap = last.rear_position - platoon_successor.position
            still_required_space = required_gap - current_gap
            LOG.trace(f"We currently have a gap of {current_gap} and thus still require {still_required_space}m")
            if last.rear_position - (required_gap + platoon_successor.length) < 0:
                # it is not possible to join because we cannot shift the current platoon successor out of the road
                LOG.warning(f"Could not make enough space to teleport vehicle {self._vid}!")
                self.in_maneuver = False
                self._platoon_role = PlatoonRole.NONE
                leader.in_maneuver = False
                self._joins_aborted += 1
                self._joins_aborted_no_space += 1
                return

            # list of vehicles that where moved already
            already_moved_vehicles = []
            # vehicle we are moving next
            current_vehicle = platoon_successor
            # we need to act to make space
            while still_required_space > 0 and current_vehicle:
                # move vehicle until min gap to its successor is reached
                LOG.trace(f"We are checking vehicle {current_vehicle.vid} for a possible move")
                # successor of the vehicle we are moving now
                current_successor = self._simulator._get_successor(current_vehicle)
                if current_vehicle is self:
                    # we cannot move the current vehicle as we are going to teleport it soon
                    # we need to select the next successor
                    LOG.trace(f"We cannot move {current_vehicle.vid} as this is the joiner!")
                    current_vehicle = current_successor
                    continue
                if current_successor is self:
                    # we cannot consider the joiner as successor, thus we need to select the next successor
                    LOG.trace("We are skipping the joiner as successor")
                    current_successor = self._simulator._get_successor(self)
                if current_successor:
                    LOG.trace(f"The successor of vehicle {current_vehicle.vid} is {current_successor.vid}")
                    # gap between vehicle we move now and its successor
                    current_back_gap = current_vehicle.rear_position - current_successor.position
                    LOG.trace(f"{current_vehicle.vid}'s back gap is {current_back_gap}m")
                    # available space we could gain during the move of the current vehicle closer to its successor (deceleration)
                    current_possible_space = current_back_gap - current_successor.min_gap
                    LOG.trace(f"We could (theoretically) gain {current_possible_space}m by moving vehicle {current_vehicle.vid}")
                else:
                    LOG.trace(f"Vehicle {current_vehicle.vid} has no successor")
                    # we have no successor
                    # thus, we can move as far as we like to
                    # but only until the end of the road
                    current_possible_space = current_vehicle.position - current_vehicle.length
                    LOG.trace(f"We could (theoretically) gain {current_possible_space}m by moving vehicle {current_vehicle.vid}")
                # space we are actually gaining during this move
                current_gained_space = min(still_required_space, current_possible_space)
                LOG.trace(f"We will gain {current_gained_space}m by moving vehicle {current_vehicle.vid}")
                # the actual gain by moving the current vehicle
                if current_gained_space > 0:
                    # we do gain
                    # move the current vehicle
                    current_vehicle._position -= current_gained_space
                    LOG.trace(f"We moved vehicle {current_vehicle.vid} to {current_vehicle.position}-{current_vehicle.rear_position} and gained {current_gained_space}m")
                    # move the previous vehicle(s) as well to keep the correct spacings between them
                    for v in already_moved_vehicles:
                        v._position -= current_gained_space
                        LOG.trace(f"We moved vehicle {v.vid} by the same distance to {v.position}-{v.rear_position}")
                    # how much space do we still need?
                    still_required_space -= current_gained_space
                else:
                    # we do not move the current vehicle and check again with its follower
                    LOG.trace(f"We can not move vehicle {current_vehicle.vid}")

                # we need to remember this vehicle
                already_moved_vehicles.append(current_vehicle)
                current_vehicle = current_successor
                LOG.trace(f"We still require {still_required_space}m")

            if still_required_space > 0:
                # we still require some space for the teleport but do not have any more vehicles to move
                # we need to abort the maneuver
                # unfortunately, we move some vehicles already but this is collateral damage?
                # NOTE: will this produce a collision as we did not move the joiner?
                LOG.warning(f"Could not make enough space to teleport vehicle {self._vid}! Aborting the join maneuver!")
                self.in_maneuver = False
                self._platoon_role = PlatoonRole.NONE
                leader.in_maneuver = False
                self._joins_aborted += 1
                self._joins_aborted_no_space += 1
                return

        # teleport the vehicle
        self._teleport(new_position, leader.lane, last.speed)

        # update the leader
        if not leader.is_in_platoon():
            # only if leader was alone before
            LOG.debug(f"{leader.vid} became a leader of platoon {leader.platoon.platoon_id}")
            leader._last_platoon_join_time = self._simulator.step
            leader._last_platoon_join_position = leader.position
            if leader._first_platoon_join_time == -1:
                # was not set before
                leader._first_platoon_join_time = self._simulator.step
                assert leader._first_platoon_join_position == -1
                # was not set before
                leader._first_platoon_join_position = leader.position
            leader._number_platoons += 1
            leader._platoon_role = PlatoonRole.LEADER
        else:
            assert leader.platoon_role is PlatoonRole.LEADER
        leader.platoon._formation.append(self)

        # update self
        self._platoon_role = PlatoonRole.FOLLOWER
        # switch to CACC
        self._cf_model = CF_Model.CACC
        self._blocked_front = False
        # set color of vehicle
        if self._simulator._gui and self._simulator.step >= self._simulator._gui_start:
            assert self._color == self.platoon.leader._color == self.color
            change_gui_vehicle_color(self._vid, leader._color)

        # collect statistics
        # the last time we joined is now
        self._last_platoon_join_time = self._simulator.step
        self._last_platoon_join_position = self._position
        if self._first_platoon_join_time == -1:
            # was not set before
            self._first_platoon_join_time = self._simulator.step
            assert self._first_platoon_join_position == -1
            # was not set before
            self._first_platoon_join_position = self._position
        self._joins_succesful += 1
        self._number_platoons += 1

        # update all members
        # update the desired speed of the platoon to the average of all platoon members
        if self._simulator._update_desired_speed:
            leader.platoon.update_desired_speed()
        leader.platoon.update_limits()
        # update formation for all members
        for vehicle in leader.platoon.formation:
            # we copy all parameters from the platoon (for now)
            # thus, the follower now drives as fast as the already existing platoon
            # (i.e., only the leader in the worst case)
            vehicle._platoon = leader.platoon
            # TODO should be replaced by call to platoon.update_cf_target_speed()
            vehicle._cf_target_speed = vehicle._platoon.desired_speed

        LOG.debug(f"{self._vid} joined platoon {leader.platoon.platoon_id} (leader: {leader.vid})")

        self.in_maneuver = False
        leader.in_maneuver = False

    def calculate_approaching_time(self, target_position: float, target_speed: float) -> float:
        """
        Calculate approximate time to approach the target position at target speed.

        Parameters
        ----------

        target_position : float
            The target position to approach
        target_speed : float
            The target speed after approaching
        """

        initial_distance = target_position - self._position
        total_approach_time = -1
        if initial_distance > 0:
            # we need to approach the platoon
            if self._speed <= target_speed:
                # we need to accelerate to approach the platoon
                self._cf_target_speed = self.max_speed
                # the time we need to accelerate to our maximum speed (given a linear acceleration)
                time_acceleration = (self.max_speed - self._speed) / self.max_acceleration
                # the time we need to decelerate to target speed (given a linear deceleration)
                time_deceleration = (self.max_speed - target_speed) / self.max_deceleration
                # the distance driven while accelerating to the maximum speed (given a linear acceleration
                distance_acceleration = (self._speed + self.max_speed) / 2 * time_acceleration
                # the distance driven while decelerating to the target speed (given a linear acceleration
                distance_deceleration = (self.max_speed + target_speed) / 2 * time_deceleration
                # the distance to drive with maximum speed (i.e., after acceleration and before deceleration)
                distance_max_speed = initial_distance - (distance_acceleration + distance_deceleration)
                # the time we need drive with the maximum speed
                time_max_speed = distance_max_speed / self.max_speed
                # the total time we need for approaching our target position
                total_approach_time = time_acceleration + time_max_speed + time_deceleration
            else:
                # we need to decelerate to approach the platoon
                # the time we need to decelerate to target speed (given a linear deceleration)
                # TODO do not use maximum deceleration for maneuver
                time_deceleration = (self._speed - target_speed) / self.max_deceleration
                # the distance driven while decelerating to the target speed (given a linear acceleration
                distance_deceleration = (self._speed + target_speed) / 2 * time_deceleration
                # the distance to drive with our current speed (i.e., before deceleration)
                distance_current_speed = initial_distance - distance_deceleration
                # the time we need drive with the current speed
                time_current_speed = distance_current_speed / self._speed
                # the total time we need for approaching our target position
                total_approach_time = time_current_speed + time_deceleration
        else:
            # we do not need to consider this case as our error is only between 0m and last.length + cacc_spacing (e.g., 9m)
            assert abs(initial_distance) <= self.length + self._cacc_spacing  # FIXME use length of last vehicle in platoon
            total_approach_time = 0  # FIXME: we ignore that for now, since the time should be very small anyhow

        return total_approach_time

    def _teleport(self, new_position: float, new_lane: int, new_speed: float):
        """
        Teleport a vehicle to a given position.

        Parameters
        ----------
        new_position : float
            The new position
        new_lane : int
            The new lane
        new_speed : int
            The new speed
        """

        current_position = self._position
        if current_position != new_position:
            self._position = new_position
            LOG.trace(f"{self._vid} teleported to {self._position} (from {current_position}, {self._position - current_position}m)")
            self._joins_teleport_position += 1
        current_lane = self._lane
        if current_lane != new_lane:
            self._lane = new_lane
            LOG.trace(f"{self._vid} switched to lane {self._lane} (from {current_lane})")
            self._joins_teleport_lane += 1
        current_speed = self._speed
        self._cf_target_speed = new_speed
        if current_speed != new_speed:
            self._speed = new_speed
            LOG.trace(f"{self._vid} changed speed to {self._speed}m/s (from {current_speed}m/s)")
            self._joins_teleport_speed += 1

        if self._simulator._record_vehicle_teleports:
            record_vehicle_teleport(
                basename=self._simulator._result_base_filename,
                step=self._simulator.step,
                vid=self._vid,
                old_position=current_position,
                old_lane=current_lane,
                old_speed=current_speed,
                new_position=new_position,
                new_lane=new_lane,
                new_speed=new_speed,
            )

    def _leave(self):
        """
        Lets a vehicle leave a platoon.

        Communication and fine-grained maneuver control is out-of-scope and thus omitted.
        """

        if self._platoon.size == 1:
            LOG.warning(f"Can not leave when driving individually ({self._vid})!")
            return
        assert self.is_in_platoon()

        leader = self._platoon.leader

        LOG.trace(f"{self._vid} is trying to leave platoon {self._platoon.platoon_id} (leader {leader.vid})")
        self._leaves_attempted += 1

        if leader.in_maneuver:
            # the platoon leader has already a (join) maneuver ongoing
            assert leader._joiner
            LOG.warning(f"{leader.vid} currently has a (join) maneuver by {leader._joiner.vid} ongoing. For now, we are going to abort this (join) maneuver as handling this situation is not yet implemented properly!")
            # TODO implement complex leave to not abort the ongoing (join) maneuver
            # abort ongoing (join) maneuver
            leader._joiner._join_approach_step = None
            leader._joiner._join_data_leader = None
            leader._joiner._join_data_last = None
            leader._joiner._join_data_new_position = None
            leader._joiner._joins_aborted += 1
            leader._joiner._joins_aborted_leave_other += 1
            leader._joiner.in_maneuver = False
            leader._joiner._platoon_role = PlatoonRole.NONE
            leader._joiner = None
            leader.in_maneuver = False
        # by checking the leader first, we should already cover cases where self == leader
        assert not self.in_maneuver

        self.in_maneuver = True
        self._platoon_role = PlatoonRole.LEAVER

        if self is leader:
            # leave at front
            self._leaves_front += 1

            if self._platoon.size == 2:
                # tell the only follower to drive individually
                follower = self._platoon.last
                assert not follower.in_maneuver
                LOG.trace(f"Only {follower.vid} is left in the platoon {self._platoon.platoon_id}. Thus, we are going to destroy the entire platoon.")
                follower._platoon_role = PlatoonRole.NONE
                follower._cf_model = CF_Model.ACC
                follower._cf_target_speed = follower._desired_speed
                follower._platoon = Platoon(follower.vid, [follower], follower._desired_speed)

                # reset color of vehicle
                if self._simulator._gui and self._simulator.step >= self._simulator._gui_start:
                    assert follower._color == follower.platoon.leader._color == follower.color
                    change_gui_vehicle_color(follower.vid, follower._color)

                # statistics
                follower._leaves_attempted += 1
                assert follower._last_platoon_join_time >= 0
                follower._time_in_platoon += follower._simulator.step - follower._last_platoon_join_time
                assert follower._last_platoon_join_position >= 0
                follower._distance_in_platoon += follower.position - follower._last_platoon_join_position
                follower._leaves_successful += 1
                follower._leaves_back += 1
            else:
                # tell the second vehicle in the platoon to become the new leader
                new_leader = self._platoon.formation[1]
                new_leader._platoon_role = PlatoonRole.LEADER
                new_leader._cf_model = CF_Model.ACC
                # platoon is changed to current leader_id
                new_leader._platoon._platoon_id = new_leader.vid

                # reset color of all remaining vehicle
                if self._simulator._gui and self._simulator.step >= self._simulator._gui_start:
                    for vehicle in self.platoon.formation[1:]:
                        change_gui_vehicle_color(vehicle.vid, new_leader._color)

                LOG.debug(f"{new_leader.vid} became leader of platoon {new_leader.platoon.platoon_id}")
        elif self is self._platoon.last:
            # leave at back
            leader.in_maneuver = True
            self._leaves_back += 1

            if self._platoon.size == 2:
                # tell the current leader to drive individually
                LOG.trace(f"Only the current leader {leader.vid} is left in the platoon {self._platoon.platoon_id}. Thus, we are going to destroy the entire platoon.")
                leader._platoon_role = PlatoonRole.NONE
                leader._cf_model = CF_Model.ACC  # TODO superfluous?
                leader._cf_target_speed = leader._desired_speed  # TODO superfluous?
                leader._platoon = Platoon(leader.vid, [leader], leader._desired_speed)

                # reset color of vehicle
                if self._simulator._gui and self._simulator.step >= self._simulator._gui_start:
                    assert leader._color == leader.platoon.leader._color == leader.color
                    change_gui_vehicle_color(leader.vid, leader._color)

                # statistics
                leader._leaves_attempted += 1
                assert leader._last_platoon_join_time >= 0
                leader._time_in_platoon += leader._simulator.step - leader._last_platoon_join_time
                assert leader._last_platoon_join_position >= 0
                leader._distance_in_platoon += leader.position - leader._last_platoon_join_position
                leader._leaves_successful += 1
                leader._leaves_front += 1
        else:
            # leave in the middle
            leader.in_maneuver = True
            self._leaves_arbitrary += 1

            # we do not need to switch lanes if we arrived
            if self._position < self._arrival_position:
                # leave the formation by changing the lane
                # TODO this looks strange in the GUI, since we do this before actually leaving the formation
                self._platoon_role = PlatoonRole.LEAVER
                self._cf_model = CF_Model.ACC

                if not self._left_lane_blocked():
                    # leave current platoon by changing to next lane
                    self._lane += 1

                    # write the lane change into related statistics
                    if self._simulator._record_vehicle_changes:
                        record_vehicle_change(
                            basename=self._simulator._result_base_filename,
                            step=self._simulator.step,
                            vid=self.vid,
                            position=self.position,
                            speed=self.speed,
                            source_lane=self.lane - 1,
                            target_lane=self.lane,
                            reason="maneuver",
                        )
                    LOG.trace(f"{self.vid} left platoon {self.platoon.platoon_id} by changing to lane {self.lane}.")

                else:
                    # the left lane is not available for leaving
                    # revert leaving process
                    self.in_maneuver = False
                    leader.in_maneuver = False
                    self._leaves_arbitrary -= 1
                    self._leaves_aborted += 1
                    self._platoon_role = PlatoonRole.FOLLOWER
                    self._cf_model = CF_Model.CACC

                    # TODO this could be just a return in future to let the leaver try again
                    #sys.exit(f"ERROR: Could not move vehicle {self._vid} to the adjacent lane to leave platoon {self.platoon.platoon_id}!")
                    LOG.trace(f"Could not move vehicle {self._vid} to the adjacent lane to leave platoon {self.platoon.platoon_id}!")
                    return

            # move all remaining platoon members further to the front
            front = self._platoon.get_front(self)
            for vehicle in self._platoon.formation[self._platoon.get_member_index(self) + 1:]:
                follower_gap = front.rear_position - vehicle.position
                gap_error = follower_gap - vehicle._cacc_spacing
                LOG.trace(f"The open gap between {front.vid} and {vehicle.vid} is {follower_gap}m (error of {gap_error}m)")
                if gap_error <= 0:
                    LOG.trace("We do not need to do anything")
                    front = vehicle
                    continue
                if vehicle.position >= vehicle.arrival_position:
                    LOG.trace(f"We are not moving {vehicle.vid} because it arrived as well")
                    continue
                LOG.trace(f"Moving follower {vehicle.vid} from {vehicle.position} by {gap_error}m ")
                vehicle._position += gap_error
                LOG.trace(f"{vehicle.vid} is now at {vehicle.position}")
                follower_gap = front.rear_position - vehicle.position
                # avoid issues due to floating point precision
                assert math.isclose(follower_gap, vehicle._cacc_spacing)
                front = vehicle
                assert self._simulator._get_predecessor(vehicle) in vehicle.platoon.formation

        # leave the platoon
        self._platoon.formation.remove(self)
        if self._simulator._update_desired_speed:
            self._platoon.update_desired_speed()
        self._platoon.update_limits()
        self._platoon.update_cf_target_speed()

        # update formation for all (remaining) members
        # TODO this could be nicer, e.g., by taking the (new) leader's updated formation
        if self._platoon.size > 1:
            # more than one remaining member
            for vehicle in self._platoon.formation:
                vehicle._platoon = self._platoon

        # leave
        LOG.debug(f"{self._vid} left platoon {self._platoon.platoon_id} (leader {self._platoon.leader.vid})")
        self._platoon_role = PlatoonRole.NONE  # the current platoon role
        self._platoon = Platoon(self._vid, [self], self._desired_speed)  # use explicit individual desired speed
        self._cf_target_speed = self._desired_speed  # we reset the cf_target_speed
        self._cf_model = CF_Model.ACC  # not necessary, but we still do it explicitly

        # reset color of vehicle
        if self._simulator._gui and self._simulator.step >= self._simulator._gui_start:
            assert self._color == self.platoon.leader._color == self.color
            change_gui_vehicle_color(self._vid, self._color)

        self.in_maneuver = False
        if self is not leader:
            leader.in_maneuver = False

        assert self._last_platoon_join_time >= 0
        self._time_in_platoon += self._simulator.step - self._last_platoon_join_time
        assert self._last_platoon_join_position >= 0
        self._distance_in_platoon += self._position - self._last_platoon_join_position
        self._leaves_successful += 1

    def _left_lane_blocked(self):
        """
        Check whether a vehicle can move to the left lane in order to leave its current platoon.
        """
        # TODO prime example for use of pandas Dataframe

        # make sure lane is within road bounds
        if self._simulator.number_of_lanes <= self._lane + 1:
            LOG.trace(
                f"ERROR: Could not move vehicle {self.vid} to the adjacent lane to leave the platoon "
                f"{self.platoon.platoon_id} because the platoon is already driving on the leftmost lane!"
            )
            return True

        # TODO same logic as in spawning and lane changing -> avoid duplicated code

        # find closest vehicle on left lane (forwards)
        closest_front = min(
            [vehicle for vehicle in self._simulator._vehicles.values() if
             vehicle.position >= self.position and vehicle.lane == (self.lane + 1)],
            key=lambda vehicle: vehicle.position,
            default=None,
        )
        # find closest vehicle on left lane (backwards)
        closest_back = max(
            [vehicle for vehicle in self._simulator._vehicles.values() if
             vehicle.position < self.position and vehicle.lane == (self.lane + 1)],
            key=lambda vehicle: vehicle.position,
            default=None,
        )
        front_gap_safe = not closest_front or is_gap_safe(
            front_position=closest_front._position,
            front_speed=closest_front.speed,
            front_max_deceleration=closest_front.max_deceleration,
            front_length=closest_front.length,
            back_position=self.position,
            back_speed=self.speed,
            back_max_acceleration=self.max_acceleration,
            back_min_gap=self.min_gap,
            step_length=self._simulator.step_length,
        )
        back_gap_safe = not closest_back or is_gap_safe(
            front_position=self.position,
            front_speed=self.speed,
            front_max_deceleration=self.max_deceleration,
            front_length=self.length,
            back_position=closest_back.position,
            back_speed=closest_back.speed,
            back_max_acceleration=closest_back.max_acceleration,
            back_min_gap=closest_back.min_gap,
            step_length=self._simulator.step_length,
        )

        return not (front_gap_safe & back_gap_safe)



#
# Copyright (c) 2020-2023 Julian Heinovski <heinovski@ccs-labs.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

import argparse
import logging
import sys
import time

import numpy as np
import pandas as pd
from tqdm import tqdm

from plafosim import CustomFormatter, __citation__, __description__, __version__
from plafosim.gui import (
    add_gui_vehicle,
    change_gui_vehicle_color,
    check_and_prepare_gui,
    close_gui,
    gui_step,
    move_gui_vehicle,
    prune_vehicles,
    start_gui,
)
from plafosim.simulator import DEFAULTS
from plafosim.util import find_resource, hex2rgb

LOG = logging.getLogger(__name__)


# TODO duplicated code with main script
def parse_args() -> argparse.Namespace:
    """
    Parse arguments given to this module.

    Returns
    -------
    argparse.Namespace
        The namespace of parsed arguments and corresponding values.
    """

    # parse some parameters
    parser = argparse.ArgumentParser(
        formatter_class=CustomFormatter,
        allow_abbrev=False,
        description=__description__,
    )

    # miscellaneous
    parser.add_argument(
        "-C", "--citation",
        action="version",
        help="show the citation information and exit",
        version=__citation__,
    )
    parser.add_argument(
        "-V", "--version",
        action="version",
        version=f"%(prog)s {__version__}",
    )
    parser.add_argument(
        "-q", "--quiet",
        action="count",
        default=0,
        help=f"The amount of verbosity levels to be removed for printing the logs to the CLI. The starting level is {logging.getLevelName(DEFAULTS['log_level'])}.",
    )
    parser.add_argument(
        "-v", "--verbosity",
        action="count",
        default=0,
        help=f"The amount of verbosity levels to be added for printing the logs to the CLI. The starting level is {logging.getLevelName(DEFAULTS['log_level'])}.",
    )

    # functionality
    parser.add_argument(
        'trace_file',
        type=str,
        help="The name of the vehicle trace file"
    )
    parser.add_argument(
        "--sumo-config",
        type=find_resource,
        default=DEFAULTS['sumo_config'],
        help="The name of the SUMO configuration file",
    )
    parser.add_argument(
        '--gui-delay',
        type=int,
        default=0,
        help="The delay used in every simulation step to visualize the current network state in ms",
    )
    parser.add_argument(
        '--track-vehicle',
        type=int,
        default=-1,
        help="The id of a vehicle to track in the GUI",
    )
    parser.add_argument(
        '--start',
        type=float,
        default=0,
        help="The first step to re-play from the trace file",
    )
    parser.add_argument(
        '--end',
        type=float,
        default=-1,
        help="The last step to re-play from the trace file. -1 is no limit",
    )
    parser.add_argument(
        '--screenshot-file',
        type=str,
        default=None,
        help="The name of the screenshot file",
    )

    # print usage without any arguments
    if len(sys.argv) < 2:
        # no argument has been passed
        print(
            parser.format_usage(),
            parser.description,
            sep='\n',
            end='',
        )
        parser.exit()

    return parser.parse_args()


def main():
    """
    The main entry point of PlaFoSim's trace replay.
    """

    args = parse_args()

    log_level = logging.getLevelName(max(DEFAULTS['log_level'] - ((args.verbosity - args.quiet) * 10), 5))
    # TODO add custom filter that prepends the log entry with the step time
    logging.basicConfig(level=log_level.upper(), format="%(levelname)s [%(name)s]: %(message)s")

    check_and_prepare_gui()

    LOG.info("Replaying vehicle trace...")

    traces = pd.read_csv(args.trace_file)

    step_length = np.diff(traces.step.unique()[:2])[0]

    start_gui(config=args.sumo_config, step_length=step_length)

    if args.track_vehicle >= 0:
        traces = traces.set_index('step')
        # limit trace to steps where tracked vehicle is present
        ego_vehicle_traces = traces[traces.id == args.track_vehicle]
        traces = traces.loc[ego_vehicle_traces.index]
        # limit trace to vehicles which are actually visible (200m radius around tracked vehicle)
        traces = traces[abs(traces.position - ego_vehicle_traces.position) <= 200]
        traces = traces.reset_index('step')
    assert not traces.empty

    min_step = max(traces.step.min(), args.start)
    max_step = min(traces.step.max(), args.end) if args.end != -1 else traces.step.max()
    LOG.debug(f"Running from {min_step}s to {max_step}s with step length {step_length}s...")

    if min_step > 0:
        gui_step(min_step)

    import traci
    for step in tqdm(np.arange(min_step, max_step, step_length), desc="Trace progress", unit='step'):
        # simulate vehicles from trace file
        for vehicle in traces[traces.step == step].itertuples():
            color = hex2rgb(vehicle.color) if 'color' in traces else (0, 255, 0)  # allow traces without color
            if str(vehicle.id) not in traci.vehicle.getIDList():
                add_gui_vehicle(
                    vehicle.id,
                    vehicle.position,
                    vehicle.lane,
                    vehicle.speed,
                    color=color,
                    track=vehicle.id == args.track_vehicle,
                )
            move_gui_vehicle(vehicle.id, vehicle.position, vehicle.lane, vehicle.speed)
            change_gui_vehicle_color(vehicle.id, color)

        # remove vehicles not in trace file (keep vehicles in trace file)
        prune_vehicles(keep_vids=list(traces.loc[traces.step == step]['id']))

        # sleep for visualization
        time.sleep(args.gui_delay / 1000)

        gui_step(step + step_length, screenshot_filename=args.screenshot_file)

    # end of file
    LOG.info("Reached end of trace file")

    # remove all vehicles
    prune_vehicles(keep_vids=[])

    close_gui()


if __name__ == "__main__":
    sys.exit(main())



#
# Copyright (c) 2020-2023 Julian Heinovski <heinovski@ccs-labs.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

import argparse
import textwrap

from plafosim.util import addLoggingLevel

__version__ = "0.17.1"

__description__ = textwrap.dedent(
    f"""\
    Platoon Formation Simulator (PlaFoSim) -- Version {__version__}.
    A simple and scalable simulator for platoon formation.

    Copyright (c) 2020-2023 Julian Heinovski <heinovski@ccs-labs.org>

    SPDX-License-Identifier: GPL-3.0-or-later

    This program comes with ABSOLUTELY NO WARRANTY.
    This is free software, and you are welcome to redistribute it under certain conditions.

    If you are working with PlaFoSim, please cite the following paper:

    Julian Heinovski, Dominik S. Buse and Falko Dressler,
    "Scalable Simulation of Platoon Formation Maneuvers with PlaFoSim,"
    Proceedings of 13th IEEE Vehicular Networking Conference (VNC 2021),
    Poster Session, Virtual Conference, November 2021, pp. 137–138.
"""
)

__citation__ = textwrap.dedent(
    """\
    @inproceedings{heinovski2021scalable,
        author = {Heinovski, Julian and Buse, Dominik S. and Dressler, Falko},
        doi = {10.1109/VNC52810.2021.9644678},
        title = {{Scalable Simulation of Platoon Formation Maneuvers with PlaFoSim}},
        pages = {137--138},
        publisher = {IEEE},
        issn = {2157-9865},
        isbn = {978-1-66544-450-7},
        address = {Virtual Conference},
        booktitle = {13th IEEE Vehicular Networking Conference (VNC 2021), Poster Session},
        month = {11},
        year = {2021},
    }
"""
)

# add additional custom log level
addLoggingLevel("TRACE", 5)


class CustomFormatter(
    argparse.ArgumentDefaultsHelpFormatter,
    argparse.RawDescriptionHelpFormatter,
    argparse.MetavarTypeHelpFormatter,
):
    """
    Metaclass combining multiple formatter classes for argparse.
    """

    pass



#
# Copyright (c) 2020-2023 Julian Heinovski <heinovski@ccs-labs.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#

from enum import Enum


class PlatoonRole(Enum):
    """
    A collection of available platoon roles.
    """

    NONE = 0  # corresponds to driving individually
    LEADER = 1  # corresponds to being the leader of a platoon
    FOLLOWER = 2  # corresponds to being a follower in a platoon
    JOINER = 3  # corresponds to being in the process of joining a platoon
    LEAVER = 4  # corresponds to being in the process of leaving a platoon



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['plafosim', 'plafosim.algorithms', 'plafosim.cli']

package_data = \
{'': ['*'], 'plafosim': ['sumocfg/*']}

install_requires = \
['numpy>=1.19,<2.0', 'ortools>=9.0,<9.1', 'pandas>=1.1,<2.0', 'tqdm>=4.61,<5.0']

entry_points = \
{'console_scripts': ['plafosim = plafosim.cli.plafosim:main',
                     'plafosim-img2video = plafosim.cli.img2video:main',
                     'plafosim-replay = plafosim.cli.trace_replay:main']}

setup_kwargs = {
    'name': 'plafosim',
    'version': '0.17.1',
    'description': 'A simple and scalable simulator for platoon formation.',
    'long_description': '# PlaFoSim - A simple and scalable simulator for platoon formation\n[![Code Version](https://img.shields.io/badge/code-v0.17.1-blue)](CHANGELOG.md)\n[![PyPI Version](https://img.shields.io/pypi/v/plafosim)](https://pypi.org/project/plafosim/)\n[![PyPI - Python Version](https://img.shields.io/pypi/pyversions/plafosim)](https://pypi.org/project/plafosim/)\n[![License](https://img.shields.io/github/license/heinovski/plafosim?color=green)](LICENSE)\n[![DOI](https://img.shields.io/badge/DOI-10.1109/VNC52810.2021.9644678-blue)](http://dx.doi.org/10.1109/VNC52810.2021.9644678)\n\n[PlaFoSim](https://www.plafosim.de) aims to facilitate and accelerate the research of platoon maneuvers and formation for individually driven vehicles.\nWhile the main focus of the simulator is on the assignment process, simulation of advertisements and maneuvers is implemented in a more abstract way.\n\nA conceptional overview of the process of Platoon Formation that is employed in PlaFoSim [1-2]:\n| Scenario | Advertisement | Assignment | Maneuver |\n| -------- | ------------- | ---------- | -------- |\n| ![Scenario](docs/scenario.png)*A new vehicle enters the highway.* | ![Advertisement](docs/advertisement.png)*The vehicle advertises itself as interested in Platooning.* | ![Assignment](docs/assignment.png)*A Vehicle-to-Platoon assignment is computed.* | ![Maneuver](docs/maneuver.png)*The new vehicle performs a join maneuver.* |\n\nPlaFoSim has been published at [IEEE VNC 2021](https://www.tkn.tu-berlin.de/bib/heinovski2021scalable/):\n\n> Julian Heinovski, Dominik S. Buse and Falko Dressler, "Scalable Simulation of Platoon Formation Maneuvers with PlaFoSim," Proceedings of 13th IEEE Vehicular Networking Conference (VNC 2021), Poster Session, Virtual Conference, November 2021, pp. 137–138.\n\n**NOTE**: PlaFoSim is (still) under heavy development.\n\n---\n\n## Installation\n\n- Install Python (>=3.7,<=3.9)\n- Install PlaFoSim from [pypi](https://pypi.org/project/plafosim/):\n```pip install plafosim```\n\n**NOTE**: The project is developed and currently only tested on Linux.\n\n## Running a Simulation\n\n**NOTE**: Since PlaFoSim is a command-line interface (CLI) application, running PlaFoSim requires using a shell.\n\n### Quickstart\n\nUse PlaFoSim\'s help to get a list of available parameters:\n\n```plafosim -h, --help```\n\nOr, run a simulation with the default configuration (`-d`):\n\n```plafosim -d```\n\n### Live GUI\n\nYou can visualize the simulation via a simple live GUI based on [SUMO](https://github.com/eclipse/sumo/), using the argument `gui` (requires installation of SUMO and decleration of `SUMO_HOME` variable):\n\n```plafosim --gui```\n\n![](docs/gui.png)\n*A screenshot of PlaFoSim\'s live GUI showing 2 platoons and various individual vehicles. Copyright © 2021 IEEE.*\n\nMore options for the live GUI can be found within the ``GUI properties`` section of the help.\n\n**NOTE**: This requires installation of SUMO (>=1.6.0) and declaration of the `SUMO_HOME` variable.\n\n### Advanced Simulation Control\n\nYou can use a variety of different parameters to customize the scenario and the simulation itself.\nE.g., use the parameter `vehicles` to configure the number of vehicles in the simulation:\n\n```plafosim --vehicles 1000```\n\nThe available parameters are grouped into different categories:\n\n```\n- road network properties\n- vehicle properties\n- trip properties\n- communication properties\n- platoon properties\n- formation properties\n- infrastructure properties\n- simulation properties\n- GUI properties\n- result recording properties\n```\n\nYou can find a list of available parameters for each category in the help:\n\n```plafosim -h, --help```\n\n#### Examples\n\n```\n# Configure a 100km freeway with ramps at every 10km\nplafosim --road-length 100 --ramp-interval 10\n\n# Configure random (normally distributed) desired driving speed of 130km/h\nplafosim --random-desired-speed true --desired-speed 36\n\n# Configure random trips for 500 vehicles\nplafosim --vehicles 500 --random-depart-position true --random-arrival-position true --depart-desired true\n\n# Pre fill the freeway with 1000 vehicles\nplafosim --vehicles 1000 --pre-fill true\n\n# Configure 50% of the vehicles with Advanced Cruise Control (ACC) and a headway time of 1.5s\nplafosim --penetration 0.5 --acc-headway-time 1.5\n\n# Enable a simple, distributed platoon formation algorithm [1] in order to form platoons every 30s\nplafosim --formation-algorithm SpeedPosition --formation-strategy distributed --execution-interval 30\n```\n\n### Faster Simulation\n\nYou can speed up the simulation performance by enabling Python\'s optimization ```PYTHONOPTIMIZE```, e.g., in order to disable assertions:\n\n```PYTHONOPTIMIZE=1 plafosim```\n\nSee the Python [documention](https://docs.python.org/3/using/cmdline.html#envvar-PYTHONOPTIMIZE) for more details.\n\n## Re-Playing a Simulation\n\nThe simulation can write a trace file including the mobility details of every simulated vehicle (default `results_vehicle_traces.csv`).\nYou can replay the simulation in the GUI (see above) based on the trace file by using the corresponding binary:\n\n```plafosim-replay results_vehicle_traces.csv```\n\nTo see all options of this script, run:\n\n```plafosim-replay -h, --help```\n\n## Extending PlaFoSim\n\nIn order to extend PlaFoSim, you first need ot install it from source.\n\n- Install [poetry](https://python-poetry.org/):\n```pip install poetry```\n- Clone this repository:\n```git clone https://github.com/heinovski/plafosim.git```\n- Navigate to newly created directory of the cloned repository in the command-line\n- Install PlaFoSim from source in editable mode:\n```poetry install```\n- Run PlaFoSim in the virtual environment:\n```poetry run plafosim```\nYou can also activate the virtual enviroment first with ```poetry shell``` and run the commands as usual (see above).\n\n### Adding a new Formation Algorithm\n\nIn order to add a new formation algorithm, you need to follow these steps:\n- Create a new sub-class of `FormationAlgorithm` (see `formation_algorithm.py`) within the sub-directory `algorithms`.\nYou can use the `Dummy` algorithm (see `algorithms/dummy.py`) as an example.\n- Add specific arguments for your algorithm to the argument parser group within the new sub-class if necessary.\n\nYou should now be able to use your new algorithm with\n```\nplafosim --formation-algorithm dummy_algorithm_name\n```\n\n## Contributing to the Project\n\nIn order to contribute, please follow these steps:\n- Install PlaFoSim from source (see Extending)\n- Make desired changes and adjust the documentation if required\n- Run the tests located in `tests` as well as the validation scripts located in `scripts` (see `.drone.yml` for details)\n- Submit a Pull Request (PR)\n\n### Documentation\n\nWhen making changes to the doce, make sure to add or adjust corresponding documentation in form of python docstrings.\nThose should use the Numpy docstring format in Sphinx and follow the [style guide](https://numpydoc.readthedocs.io/en/latest/format.html) by numpydoc to ensure consistency and compatibility.\n\n### Testing\n\nWhen adding methods and functions, make sure to add corresponding unit tests for `py.test`.\nThe tests are located under `tests` and can be executed with `./scripts/run-pytest.sh`.\nThis will also generate a test coverage report.\n\n### Validation\n\nTo validate the behavior of PlaFoSim, it is compared to SUMO 1.6.0 by means of simulation results (e.g., vehicle traces).\nThe corresponding scripts are located under `scripts` and executed withn CI/CD pipelines.\nYou can have a look at `.drone.yml` for details regarding the execution.\n\n### Profiling\n\nYou can profile the runtime of PlaFoSim\'s code by using [cProfile](https://docs.python.org/3/library/profile.html#module-cProfile):\n\n```poetry run python -m cProfile -o profile.out -m plafosim.cli.plafosim```\n\nYou can visualize the results of the profiling run by using [SnakeViz](https://jiffyclub.github.io/snakeviz/):\n\n```snakeviz profile.out```\n\n## Contributors & Citing\n\nPlaFoSim was originally designed and built by [Julian Heinovski](https://github.com/heinovski/) with the help of [Dominik S. Buse](https://github.com/dbuse/).\nIt is currently maintained by [Julian Heinovski](https://github.com/heinovski/).\nThe list of all authors can be found [here](AUTHORS.md).\n\nIf you are working with `PlaFoSim`, we would appreciate a citation of [our paper](https://www.tkn.tu-berlin.de/bib/heinovski2021scalable/):\n\n> Julian Heinovski, Dominik S. Buse and Falko Dressler, "Scalable Simulation of Platoon Formation Maneuvers with PlaFoSim," Proceedings of 13th IEEE Vehicular Networking Conference (VNC 2021), Poster Session, Virtual Conference, November 2021, pp. 137–138.\n\n```bibtex\n@inproceedings{heinovski2021scalable,\n    author = {Heinovski, Julian and Buse, Dominik S. and Dressler, Falko},\n    doi = {10.1109/VNC52810.2021.9644678},\n    title = {{Scalable Simulation of Platoon Formation Maneuvers with PlaFoSim}},\n    pages = {137--138},\n    publisher = {IEEE},\n    issn = {2157-9865},\n    isbn = {978-1-66544-450-7},\n    address = {Virtual Conference},\n    booktitle = {13th IEEE Vehicular Networking Conference (VNC 2021), Poster Session},\n    month = {11},\n    year = {2021},\n}\n```\n\n## License\n\nPlaFoSim is licensed under the terms of the GNU General Public License 3.0 or later.\n\n```\n# Copyright (c) 2020-2023 Julian Heinovski <heinovski@ccs-labs.org>\n#\n# SPDX-License-Identifier: GPL-3.0-or-later\n#\n# This program is free software: you can redistribute it and/or modify\n# it under the terms of the GNU General Public License as published by\n# the Free Software Foundation, either version 3 of the License, or\n# any later version.\n#\n# This program is distributed in the hope that it will be useful,\n# but WITHOUT ANY WARRANTY; without even the implied warranty of\n# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n# GNU General Public License for more details.\n#\n# You should have received a copy of the GNU General Public License\n# along with this program.  If not, see <https://www.gnu.org/licenses/>.\n```\n\n## List of References\n\n[1] Julian Heinovski and Falko Dressler, ["Platoon Formation: Optimized Car to Platoon Assignment Strategies and Protocols,"](https://www.tkn.tu-berlin.de/bib/heinovski2018platoon/) Proceedings of 10th IEEE Vehicular Networking Conference (VNC 2018), Taipei, Taiwan, December 2018.\n\n[2] Julian Heinovski, ["Platoon Formation: Car-to-Platoon Assignments for Individual Cars,"](https://www.tkn.tu-berlin.de/bib/heinovski2019platoon/) Proceedings of International Conference on Networked Systems (NetSys 2019), PhD Forum, Munich, Germany, March 2019.\n',
    'author': 'Julian Heinovski',
    'author_email': 'heinovski@ccs-labs.org',
    'maintainer': 'Julian Heinovski',
    'maintainer_email': 'heinovski@ccs-labs.org',
    'url': 'https://www.plafosim.de',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<3.10',
}


setup(**setup_kwargs)
